"""Methods to evaluate pairs."""
from pathlib import Path
import inspect
from typing import Generator, Any, Literal
from concurrent.futures import ProcessPoolExecutor
import itertools
import functools

import polars as pl
import pandas as pd
import numpy as np
import numpy.typing as npt
from numba import float64, guvectorize
from arch.bootstrap import StationaryBootstrap, optimal_block_length

from nwm_explorer.nwm import ModelConfiguration
from nwm_explorer.pairs import scan_pairs, GROUP_SPECIFICATIONS
from nwm_explorer.logger import get_logger
from nwm_explorer.routelink import download_routelink
from nwm_explorer.constants import (Metric, SUBDIRECTORIES, LRU_CACHE_SIZES,
    METRIC_SIGNIFICANCE_THRESHOLD, MetricFunction)

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def nash_sutcliffe_efficiency(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba compatible implementation of Nash-Sutcliffe Model Efficiency.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations,
        measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations, forecasts, or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    variance = np.sum((y_true - np.mean(y_true)) ** 2.0)
    if variance == 0:
        result[0] = np.nan
        return
    result[0] = 1.0 - np.sum((y_true - y_pred) ** 2.0) / variance

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def relative_mean_bias(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba compatible implementation of signed mean relative bias.
    Also called mean relative error or fractional bias.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations,
        measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    total = np.sum(y_true)
    if total == 0.0:
        result[0] = np.nan
        return
    result[0] = np.sum(y_pred - y_true) / np.sum(y_true)

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def pearson_correlation_coefficient(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba compatible implementation of the Pearson correlation
    coefficient.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations,
        measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    y_true_dev = y_true - np.mean(y_true)
    y_pred_dev = y_pred - np.mean(y_pred)
    num = np.sum(y_true_dev * y_pred_dev)
    den = (
        np.sqrt(np.sum(y_true_dev ** 2)) *
        np.sqrt(np.sum(y_pred_dev ** 2))
        )
    if den == 0:
        result[0] = np.nan
        return
    result[0] = num / den

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def relative_standard_deviation(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba compatible implementation of relative standard deviation,
    required to compute Kling-Gupta Model Efficiency.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations,
        measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    std_dev = np.std(y_true)
    if std_dev == 0:
        result[0] = np.nan
        return
    result[0] = np.std(y_pred) / std_dev

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def relative_mean(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba compatible implementation of relative mean,
    required to compute Kling-Gupta Model Efficiency.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations,
        measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    mean = np.mean(y_true)
    if mean == 0:
        result[0] = np.nan
        return
    result[0] = np.mean(y_pred) / mean

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def relative_median(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba compatible implementation of relative mean,
    required to compute Kling-Gupta Model Efficiency.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations,
        measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    median = np.median(y_true)
    if median == 0:
        result[0] = np.nan
        return
    result[0] = np.median(y_pred) / median

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def relative_minimum(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba compatible implementation of relative minimum.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations,
        measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    minimum = np.min(y_true)
    if minimum == 0:
        result[0] = np.nan
        return
    result[0] = np.min(y_pred) / minimum

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def relative_maximum(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba compatible implementation of relative maximum.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations,
        measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    maximum = np.max(y_true)
    if maximum == 0:
        result[0] = np.nan
        return
    result[0] = np.max(y_pred) / maximum

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def kling_gupta_efficiency(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba compatible implementation of Kling-Gupta Model Efficiency.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations,
        measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    correlation = np.empty(shape=1, dtype=np.float64)
    pearson_correlation_coefficient(y_true, y_pred, correlation)
    rel_var = np.empty(shape=1, dtype=np.float64)
    relative_standard_deviation(y_true, y_pred, rel_var)
    rel_mean = np.empty(shape=1, dtype=np.float64)
    relative_mean(y_true, y_pred, rel_mean)
    result[0] = (1.0 - np.sqrt(
        ((correlation[0] - 1.0)) ** 2.0 +
        ((rel_var[0] - 1.0)) ** 2.0 +
        ((rel_mean[0] - 1.0)) ** 2.0
        ))

METRIC_FUNCTIONS: dict[Metric, MetricFunction] = {
    Metric.NASH_SUTCLIFFE_EFFICIENCY: nash_sutcliffe_efficiency,
    Metric.RELATIVE_MEAN_BIAS: relative_mean_bias,
    Metric.PEARSON_CORRELATION_COEFFICIENT: pearson_correlation_coefficient,
    Metric.RELATIVE_MEAN: relative_mean,
    Metric.RELATIVE_MEDIAN: relative_median,
    Metric.RELATIVE_MINIMUM: relative_minimum,
    Metric.RELATIVE_MAXIMUM: relative_maximum,
    Metric.RELATIVE_STANDARD_DEVIATION: relative_standard_deviation,
    Metric.KLING_GUPTA_EFFICIENCY: kling_gupta_efficiency
}
"""Mapping from metrics to functions used to compute them."""

def bootstrap_metrics(
    data: pd.DataFrame,
    minimum_sample_size: int = 30,
    minimum_mean: float = 0.01,
    minimum_variance: float = 0.000025
    ) -> dict[str, Any]:
    """
    Use stationary bootstrap to generate metrics with confidence intervals.
    Returns a dictionary. Assumed use is as record for pandas.DataFrame.from_records.

    Parameters
    ----------
    data: pandas.DataFrame, required
        DataFrame of pairs for a specific feature and lead time.
    minimum_sample_size: int, optional, default 30
        Minimum number of samples required to compute confidence intervals.
    minimum_mean: float, optional, default 0.01
        Smallest mean value of observed time series required to compute
        confidence intervals.
    minimum_variance: float, optional, default 0.000025
        Smallest variance of observed time series required to compute
        confidence intervals.
    
    Returns
    -------
    dict[str, Any]
    """
    # Start building results
    result = {
        "nwm_feature_id": data["nwm_feature_id"].iloc[0],
        "lead_time_hours_min": data["lead_time_hours_min"].min(),
        "predicted_cfs_min": data["predicted_cfs_min"].min(),
        "predicted_cfs_median": data["predicted_cfs_median"].median(),
        "predicted_cfs_max": data["predicted_cfs_max"].max(),
        "predicted_value_time_min": data["predicted_value_time_min"].min(),
        "predicted_value_time_max": data["predicted_value_time_max"].max(),
        "reference_time_min": data["reference_time_min"].min(),
        "reference_time_max": data["reference_time_max"].max(),
        "usgs_site_code": data["usgs_site_code"].iloc[0],
        "observed_cfs_min": data["observed_cfs_min"].min(),
        "observed_cfs_median": data["observed_cfs_median"].median(),
        "observed_cfs_max": data["observed_cfs_max"].max(),
        "observed_value_time_min": data["observed_value_time_min"].min(),
        "observed_value_time_max": data["observed_value_time_max"].max(),
        "sample_size": data["nwm_feature_id"].count()
    }

    # Compute metrics
    for rank in ["min", "median", "max"]:
        # Extract numpy arrays for Numba functions
        y_true = data[f"observed_cfs_{rank}"].to_numpy(dtype=np.float64)
        y_pred = data[f"predicted_cfs_{rank}"].to_numpy(dtype=np.float64)

        # Compute each metric
        for label, func in METRIC_FUNCTIONS.items():
            # Point estimate
            # NOTE Numba has magic that implicitly instantiates point, but
            #  it makes the linter mad.
            point = np.empty(shape=1, dtype=np.float64)
            func(y_true, y_pred, point)
            result[f"{label}_{rank}_point"] = point[0]

            # Sample size too small reliably to compute confidence interval
            if result["sample_size"] < minimum_sample_size:
                result[f"{label}_{rank}_lower"] = np.nan
                result[f"{label}_{rank}_upper"] = np.nan
                continue

            # Values too small to reliably compute confidence interval
            if np.mean(y_true) < minimum_mean:
                result[f"{label}_{rank}_lower"] = np.nan
                result[f"{label}_{rank}_upper"] = np.nan
                continue

            # Variance too small to reliably compute confidence interval
            if np.var(y_true) < minimum_variance:
                result[f"{label}_{rank}_lower"] = np.nan
                result[f"{label}_{rank}_upper"] = np.nan
                continue

            # Optimal block size
            # NOTE Normalizing the values seems to produce more consistent
            #  block sizes. Here we let the "true" values determine the block size.
            max_value = np.max(y_true) * 1.01
            normalized = y_true / max_value
            block_size = optimal_block_length(normalized)["stationary"][0]
            if np.isnan(block_size):
                block_size = 1
            else:
                block_size = max(1, int(block_size))

            # Resample the array index to apply to both y_true and y_pred
            index = np.arange(y_true.size)
            bs = StationaryBootstrap(
                block_size,
                index,
                seed=2025
                )

            # Generate posterior distribution
            posterior = []
            estimate = np.empty(shape=1, dtype=np.float64)
            for samples in bs.bootstrap(1000):
                # Generate an index
                idx = samples[0][0]

                # Apply the new index and compute the metric
                func(y_true[idx], y_pred[idx], estimate)
                posterior.append(estimate[0])

            # Compute confidence interval
            ci = np.quantile(posterior, [0.025, 0.975])
            result[f"{label}_{rank}_lower"] = ci[0]
            result[f"{label}_{rank}_upper"] = ci[1]
    return result

def load_pool(
        root: Path,
        configuration: ModelConfiguration,
        start_time: pd.Timestamp,
        end_time: pd.Timestamp,
        lead_time_interval: int,
        features: npt.NDArray[np.int64]
        ) -> pl.DataFrame:
    """
    Load and group forecast pairs into lead time pools.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    configuration: ModelConfiguration
        NWM Model configuration.
    start_time: pandas.Timestamp
        First reference time.
    end_time: pandas.Timestamp
        Last reference time.
    lead_time_interval: int
        Lead time scale to aggregate over in hours.
    features: numpy.ndarray[int64], required
        Array of channel features to retrieve from parquet store.
    
    Returns
    -------
    polars.DataFrame
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # Load data
    logger.info("Loading pairs")
    dataframes = []
    for m in pd.date_range(start_time, end_time, freq="1MS"):
        dataframes.append(
            scan_pairs(root, cache=True).filter(
                pl.col("configuration") == configuration,
                pl.col("year") == m.year,
                pl.col("month") == m.month,
                pl.col("reference_time") >= start_time,
                pl.col("reference_time") <= end_time,
                pl.col("nwm_feature_id").is_in(features)
            ).collect()
        )

    # Compute lead time and pool pairs
    logger.info("Pooling pairs")
    return pl.concat(
        dataframes
    ).with_columns(
        lead_time_hours_min=(pl.col("predicted_value_time_min").sub(pl.col("reference_time")) /
            pl.duration(hours=lead_time_interval)).floor().cast(pl.Int32).mul(lead_time_interval)
    ).sort([
        "configuration",
        "nwm_feature_id",
        "lead_time_hours_min",
        "value_time"]
    ).group_by_dynamic(
        "value_time",
        every=f"{lead_time_interval}h",
        group_by=["configuration", "nwm_feature_id", "lead_time_hours_min"]
    ).agg(
        pl.col("predicted_cfs_min").min(),
        pl.col("predicted_cfs_median").median(),
        pl.col("predicted_cfs_max").max(),
        pl.col("predicted_value_time_min").min(),
        pl.col("predicted_value_time_max").max(),
        pl.col("reference_time").min().alias("reference_time_min"),
        pl.col("reference_time").max().alias("reference_time_max"),
        pl.col("usgs_site_code").first(),
        pl.col("observed_cfs_min").min(),
        pl.col("observed_cfs_median").median(),
        pl.col("observed_cfs_max").max(),
        pl.col("observed_value_time_min").min(),
        pl.col("observed_value_time_max").max()
    )

def prediction_pool_generator(
        root: Path,
        configuration: ModelConfiguration,
        start_time: pd.Timestamp,
        end_time: pd.Timestamp,
        lead_time_interval: int,
        sites_per_chunk: int = 1
) -> Generator[list[pd.DataFrame]]:
    """
    Iteratively, load and group forecast pairs into lead time pools. Returns a
    list of DataFrame for each nwm_feature_id and lead_time_hours_min combination.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    configuration: ModelConfiguration
        NWM Model configuration.
    start_time: pandas.Timestamp
        First reference time.
    end_time: pandas.Timestamp
        Last reference time.
    lead_time_interval: int
        Lead time scale to aggregate over in hours.
    sites_per_chunk: int, optional, default 1
        Maximum number of sites returned per iteration.

    Returns
    -------
    Generator[list[pandas.DataFrame]]
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # Generate feature list
    feature_list = download_routelink(
        root
    ).select(
        "nwm_feature_id"
    ).collect()["nwm_feature_id"].to_numpy()

    # Feature chunks
    number_of_chunks = len(feature_list) // sites_per_chunk + 1
    feature_chunks = np.array_split(feature_list, number_of_chunks)

    # Yield pool generator
    for features in feature_chunks:
        # Load data
        data = load_pool(
            root=root,
            configuration=configuration,
            start_time=start_time,
            end_time=end_time,
            lead_time_interval=lead_time_interval,
            features=features
        )

        # Check for data
        if data.is_empty():
            logger.debug("Empty pool, trying again")
            continue

        # Generate groups
        yield [df.to_pandas() for _, df in data.group_by(["nwm_feature_id", "lead_time_hours_min"])]

def evaluate(
        label: str,
        root: Path,
        start_time: pd.Timestamp,
        end_time: pd.Timestamp,
        processes: int = 1,
        sites_per_chunk: int = 1
) -> None:
    """
    Iteratively, load and group forecast pairs into lead time pools. Returns a
    list of DataFrame for each nwm_feature_id and lead_time_hours_min combination.

    Parameters
    ----------
    label: str, required
        Machine-friendly label used to generate parquet store.
    root: pathlib.Path
        Root data directory.
    start_time: pandas.Timestamp
        First reference time.
    end_time: pandas.Timestamp
        Last reference time.
    processes: int, optional, default 1
        Number of parallel processes to use for computation.
    sites_per_chunk: int, optional, default 1
        Maximum number of sites to load into memory at once.

    Returns
    -------
    Generator[list[pandas.DataFrame]]
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # Logging
    logger.info("Running evaluation %s", label)
    logger.info("Range: %s to %s", start_time, end_time)

    # Start process pool
    logger.info("Starting %d processes", processes)
    subdirectory = SUBDIRECTORIES["evaluations"]
    with ProcessPoolExecutor(max_workers=processes) as parallel_computer:
        # Process each configuration
        for config, specs in GROUP_SPECIFICATIONS.items():
            # Prepare output file
            ofile = root / f"{subdirectory}/label={label}/configuration={config}/E0.parquet"
            if ofile.exists():
                logger.info("Found %s", ofile)
                continue
            ofile.parent.mkdir(exist_ok=True, parents=True)

            # Process in chunks
            logger.info("Evaluating %s", config)
            logger.info("Grouping into chunks of %d sites", sites_per_chunk)
            batch_counter = itertools.count(1)
            dataframes = []
            for groups in prediction_pool_generator(
                root=root,
                configuration=config,
                start_time=start_time,
                end_time=end_time,
                lead_time_interval=specs.window_interval,
                sites_per_chunk=sites_per_chunk
            ):
                # Chunk size
                logger.info("Batch: %d", next(batch_counter))
                chunksize = len(groups) // processes + 1
                logger.info("Evaluating %d groups", len(groups))
                logger.info("Running %d groups per process", chunksize)

                # Compute and collect results
                logger.info("Computing metrics")
                dataframes.append(
                    pd.DataFrame.from_records(
                        parallel_computer.map(
                            bootstrap_metrics, groups, chunksize=chunksize
                        )
                    )
                )

            # Concatenate into a single dataframe
            logger.info("Concatenating groups")
            results = pd.concat(dataframes, ignore_index=True)

            # Save results
            logger.info("Saving %s", ofile)
            pl.DataFrame(results).write_parquet(ofile)

def scan_evaluations_no_cache(root: Path) -> pl.LazyFrame:
    """
    Return polars.LazyFrame of Evaluation results.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.

    Returns
    -------
    polars.LazyFrame
    """
    subdirectory = SUBDIRECTORIES["evaluations"]
    return pl.scan_parquet(
        root / f"{subdirectory}/",
        hive_schema={
            "label": pl.String,
            "configuration": pl.Enum(ModelConfiguration)
        }
    )

@functools.lru_cache(LRU_CACHE_SIZES["evaluations"])
def scan_evaluations_cache(root: Path) -> pl.LazyFrame:
    """
    Return polars.LazyFrame of Evaluation results. Cache LazyFrame.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.

    Returns
    -------
    polars.LazyFrame
    """
    return scan_evaluations_no_cache(root)

def scan_evaluations(root: Path, cache: bool = False) -> pl.LazyFrame:
    """
    Return polars.LazyFrame of Evaluation results. Optionally, cache LazyFrame.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    cache: bool, optional, default False
        If True, caches LazyFrame.

    Returns
    -------
    polars.LazyFrame
    """
    if cache:
        return scan_evaluations_cache(root)
    return scan_evaluations_no_cache(root)

def load_metrics_no_cache(
        root: Path,
        label: str,
        configuration: ModelConfiguration,
        metric: Metric,
        lead_time_hours_min: int = 0,
        rank: Literal["min", "median", "max"] = "median",
        additional_columns: tuple[str] | None = None,
        significant: bool = False,
) -> pl.DataFrame:
    """
    Returns DataFrame of metrics.

    Parameters
    ----------
    root: pathlib.Path
        Root of data directory.
    label: str
        Label of evaluation.
    configuration: ModelConfiguration
        National Water Model configuration.
    metric: Metric
        Evaluation metric.
    lead_time_hours_min: int, optional, default 0
        Minimum lead time in hours.
    rank: Literal["min", "median", "max"], optional, default "median"
        Aggregated function applied to predictions and observations. Varies by
        configuration. Analysis & Assimilation and Medium Range configurations
        will typically return daily streamflow aggregated to minimum, median, or
        maximum streamflow. Short Range CONUS, Hawaii, and Puerto Rico return
        6-hourly aggregation. Short Range Alaska returns 5-hourly aggregations.
    additional_columns: tuple[str], optional, default ("nwm_feature_id",)
        Additional columns (often metadata) to return with metric values.
    significant: bool, optional, default False
        If True, only return 'statistically significant' values.
    
    Returns
    -------
    polars.DataFrame
    """
    # Check additional columns
    if additional_columns is None:
        additional_columns = ["nwm_feature_id"]
    else:
        additional_columns = list(additional_columns)

    # Retrieve
    data = scan_evaluations(
        root
    ).filter(
        pl.col("label") == label,
        pl.col("configuration") == configuration,
        pl.col("lead_time_hours_min") == lead_time_hours_min
    ).select(
        [
            f"{metric}_{rank}_lower",
            f"{metric}_{rank}_point",
            f"{metric}_{rank}_upper"
        ] + additional_columns
    ).collect()

    # Apply significance test
    if significant:
        threshold = METRIC_SIGNIFICANCE_THRESHOLD[metric]
        return data.with_columns(
            significant=~(
                (pl.col(f"{metric}_{rank}_lower") <= threshold) &
                (pl.col(f"{metric}_{rank}_upper") >= threshold)
            )
        ).filter(pl.col("significant"))
    return data

@functools.lru_cache(LRU_CACHE_SIZES["evaluations"])
def load_metrics_cache(
        root: Path,
        label: str,
        configuration: ModelConfiguration,
        metric: Metric,
        lead_time_hours_min: int = 0,
        rank: Literal["min", "median", "max"] = "median",
        additional_columns: tuple[str] | None = None,
        significant: bool = False
) -> pl.DataFrame:
    """
    Returns DataFrame of metrics. Cache result.

    Parameters
    ----------
    root: pathlib.Path
        Root of data directory.
    label: str
        Label of evaluation.
    configuration: ModelConfiguration
        National Water Model configuration.
    metric: Metric
        Evaluation metric.
    lead_time_hours_min: int, optional, default 0
        Minimum lead time in hours.
    rank: Literal["min", "median", "max"], optional, default "median"
        Aggregated function applied to predictions and observations. Varies by
        configuration. Analysis & Assimilation and Medium Range configurations
        will typically return daily streamflow aggregated to minimum, median, or
        maximum streamflow. Short Range CONUS, Hawaii, and Puerto Rico return
        6-hourly aggregation. Short Range Alaska returns 5-hourly aggregations.
    additional_columns: tuple[str], optional, default ("nwm_feature_id",)
        Additional columns (often metadata) to return with metric values.
    significant: bool, optional, default False
        If True, only return 'statistically significant' values.
    
    Returns
    -------
    polars.DataFrame
    """
    # Check additional columns
    if additional_columns is None:
        additional_columns = ["nwm_feature_id"]
    else:
        additional_columns = list(additional_columns)

    # Retrieve
    data = scan_evaluations(
        root, cache=True
    ).filter(
        pl.col("label") == label,
        pl.col("configuration") == configuration,
        pl.col("lead_time_hours_min") == lead_time_hours_min
    ).select(
        [
            f"{metric}_{rank}_lower",
            f"{metric}_{rank}_point",
            f"{metric}_{rank}_upper"
        ] + additional_columns
    ).collect()

    # Apply significance test
    if significant:
        threshold = METRIC_SIGNIFICANCE_THRESHOLD[metric]
        return data.with_columns(
            significant=~(
                (pl.col(f"{metric}_{rank}_lower") <= threshold) &
                (pl.col(f"{metric}_{rank}_upper") >= threshold)
            )
        ).filter(pl.col("significant"))
    return data

def load_metrics(
        root: Path,
        label: str,
        configuration: ModelConfiguration,
        metric: Metric,
        lead_time_hours_min: int = 0,
        rank: Literal["min", "median", "max"] = "median",
        additional_columns: tuple[str] | None = None,
        significant: bool = False,
        cache: bool = False
) -> pl.DataFrame:
    """
    Returns DataFrame of metrics. Cache result.

    Parameters
    ----------
    root: pathlib.Path
        Root of data directory.
    label: str
        Label of evaluation.
    configuration: ModelConfiguration
        National Water Model configuration.
    metric: Metric
        Evaluation metric.
    lead_time_hours_min: int, optional, default 0
        Minimum lead time in hours.
    rank: Literal["min", "median", "max"], optional, default "median"
        Aggregated function applied to predictions and observations. Varies by
        configuration. Analysis & Assimilation and Medium Range configurations
        will typically return daily streamflow aggregated to minimum, median, or
        maximum streamflow. Short Range CONUS, Hawaii, and Puerto Rico return
        6-hourly aggregation. Short Range Alaska returns 5-hourly aggregations.
    additional_columns: tuple[str], optional, default ("nwm_feature_id",)
        Additional columns (often metadata) to return with metric values.
    significant: bool, optional, default False
        If True, only return 'statistically significant' values.
    cache: bool, optional, default False
        If true, cache result.
    
    Returns
    -------
    polars.DataFrame
    """
    # Retrieve
    if cache:
        return load_metrics_cache(
            root=root,
            label=label,
            configuration=configuration,
            metric=metric,
            lead_time_hours_min=lead_time_hours_min,
            rank=rank,
            additional_columns=additional_columns,
            significant=significant
        )
    return load_metrics_no_cache(
        root=root,
        label=label,
        configuration=configuration,
        metric=metric,
        lead_time_hours_min=lead_time_hours_min,
        rank=rank,
        additional_columns=additional_columns,
        significant=significant
    )

def load_site_metrics(
        root: Path,
        label: str,
        configuration: ModelConfiguration,
        metric: Metric,
        nwm_feature_id: int,
        rank: Literal["min", "median", "max"] = "median",
        cache: bool = False
) -> pl.DataFrame:
    """
    Returns DataFrame of metrics for a single site.

    Parameters
    ----------
    root: pathlib.Path
        Root of data directory.
    label: str
        Label of evaluation.
    configuration: ModelConfiguration
        National Water Model configuration.
    metric: Metric
        Evaluation metric.
    nwm_feature_id: int
        National Water Model channel feature identifier.
    rank: Literal["min", "median", "max"], optional, default "median"
        Aggregated function applied to predictions and observations. Varies by
        configuration. Analysis & Assimilation and Medium Range configurations
        will typically return daily streamflow aggregated to minimum, median, or
        maximum streamflow. Short Range CONUS, Hawaii, and Puerto Rico return
        6-hourly aggregation. Short Range Alaska returns 5-hourly aggregations.
    cache: bool, optional, default False
        If true, cache underlying polars.LazyFrame.
    
    Returns
    -------
    polars.DataFrame
    """
    # Retrieve
    return scan_evaluations(root, cache).filter(
        pl.col("label") == label,
        pl.col("configuration") == configuration,
        pl.col("nwm_feature_id") == nwm_feature_id
    ).select(
        [
            "lead_time_hours_min",
            f"{metric}_{rank}_lower",
            f"{metric}_{rank}_point",
            f"{metric}_{rank}_upper"
        ]
    ).collect().sort("lead_time_hours_min")



"""
Download and process USGS streamflow observations.

Methods
-------
- download_site_table
- download_usgs
- scan_usgs
"""
from pathlib import Path
import inspect
import json
from dataclasses import dataclass
from tempfile import TemporaryDirectory
from time import sleep
import functools
from typing import Generator

import us
import pandas as pd
import polars as pl
import geopandas as gpd

from nwm_explorer.logger import get_logger
from nwm_explorer.downloads import download_files
from nwm_explorer.configuration import Configuration
from nwm_explorer.constants import (LRU_CACHE_SIZES, SUBDIRECTORIES, SiteTypeSlug,
    NWIS_BASE_URL, MONITORING_LOCATION_BASE_URL, STATE_LIST, SITE_TYPES,
    SITE_SCHEMA)

def download_site_table(
        root: Path,
        config: Configuration
    ):
    """
    Download and process NWM output.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    config: Configuration
        Configuration object.
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # Output directory
    odir = root / SUBDIRECTORIES["site_table"]
    odir.mkdir(exist_ok=True, parents=True)

    # Build site table
    logger.info("Building site table")
    for state in STATE_LIST:
        for site_type in SITE_TYPES:
            # Build state file
            ofile = odir / f"site_type_slug={site_type.slug}" / (state.abbr.lower() + ".parquet")
            if ofile.exists():
                logger.info("Found %s", ofile)
                continue
            logger.info("Buildling %s", ofile)
            ofile.parent.mkdir(exist_ok=True, parents=True)

            # Build URL
            url = (MONITORING_LOCATION_BASE_URL + site_type.code +
                f"&state_code={state.fips}" + f"&api_key={config.usgs_api_key}")
            logger.info("Downloading %s", url)

            # Fetch
            for attempt in range(3):
                try:
                    gdf = gpd.read_file(url)
                    break
                except Exception as e:
                    logger.info(e)
                    gdf = None
                    sleep(2 ** attempt)

            # Check for data
            if gdf is None:
                logger.info("Unable to retrieve")
                continue
            if gdf.empty:
                logger.info("Empty response")
                continue

            # Convert to polars
            gdf["longitude"] = gdf.geometry.x
            gdf["latitude"] = gdf.geometry.y
            data = pl.DataFrame(
                gdf.drop("geometry", axis=1),
                schema_overrides=SITE_SCHEMA
            )

            # Save
            logger.info("Saving %s", ofile)
            data.write_parquet(ofile)

def expand_site_table(
        ifile: Path,
        ofile: Path
) -> None:
    """
    Manually expand the site.

    Parameters
    ----------
    ifile: pathlib.Path
        GeoJSON file of USGS monitoring locations. (e.g.
        Path('new_york_sites.json'))
    ofile: pathlib.Path
        New parquet file to add to the application site table. (e.g.
        Path('site_table/site_type_slug=stream/ny_2.parquet'))
    
    Returns
    -------
    None
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # Open file
    logger.info("Importing %s", ifile)
    gdf = gpd.read_file(ifile)

    # Convert to polars
    logger.info("Converting to parquet")
    gdf["longitude"] = gdf.geometry.x
    gdf["latitude"] = gdf.geometry.y
    data = pl.DataFrame(
        gdf.drop("geometry", axis=1),
        schema_overrides=SITE_SCHEMA
    )

    # Save
    logger.info("Saving %s", ofile)
    data.write_parquet(ofile)

@functools.lru_cache(LRU_CACHE_SIZES["usgs"])
def scan_site_table(root: Path) -> pl.LazyFrame:
    """
    Return polars.LazyFrame of USGS sites.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    """
    return pl.scan_parquet(
        root / SUBDIRECTORIES["site_table"],
        hive_schema={"site_type_slug": pl.Enum(SiteTypeSlug)}
        )

def load_site_information(
        root: Path,
        usgs_site_code: str,
        columns: list[str] | None = None,
        rename: dict[str, str] | None = None
        ) -> pl.DataFrame:
    """
    Return site specific information.

    Parameters
    ----------
    root: pathlib.Path
        Root directory containing site information tables.
    usgs_site_code: str
        USGS site code.
    columns: list[str], optional
        List of columns to return. Default: ["id", "site_type", "drainage_area",
        "hydrologic_unit_code", "contributing_drainage_area", "monitoring_location_name"]
    rename: dict[str, str], optional
        Optional mapping to rename columns.
    
    Returns
    -------
    pl.DataFrame
    """
    # Set columns
    if columns is None:
        columns = [
            "monitoring_location_name",
            "monitoring_location_number",
            "hydrologic_unit_code",
            "site_type",
            "drainage_area",
            "contributing_drainage_area"
        ]

    # Get data
    if "monitoring_location_number" in columns:
        df = scan_site_table(root).select(columns).filter(
            pl.col("monitoring_location_number") == usgs_site_code
        ).collect()
    else:
        df = scan_site_table(root).select(columns+["monitoring_location_number"]).filter(
            pl.col("monitoring_location_number") == usgs_site_code
        ).collect().select(columns)

    # Rename columns
    if rename is not None:
        df = df.rename(rename)
    return df

@functools.lru_cache(LRU_CACHE_SIZES["usgs"])
def lookup_site_state_code_cache(root: Path, usgs_site_code: str) -> str:
    """
    Given a USGS site code, return the US state as a lower case two-character
    abbreviation.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    usgs_site_code: str
        USGS site code.
    
    Returns
    -------
    str
    """
    fips = scan_site_table(root).filter(
        pl.col("monitoring_location_number") == usgs_site_code
    ).select("state_code").collect()["state_code"].item(0)
    return us.states.lookup(fips).abbr.lower()

def lookup_site_state_code(root: Path, usgs_site_code: str) -> str:
    """
    Given a USGS site code, return the US state as a lower case two-character
    abbreviation.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    usgs_site_code: str
        USGS site code.
    
    Returns
    -------
    str
    """
    return lookup_site_state_code_cache(root, usgs_site_code)

def json_validator(ifile: Path) -> None:
    """
    Open and read json file data.
    """
    with ifile.open("r") as fi:
        json.loads(fi.read())

def build_usgs_filepath(
        root: Path,
        state_code: str,
        date: pd.Timestamp
) -> Path:
    """Build and return USGS hive-partitioned parquet file path."""
    # Build path
    state = f"state_code={state_code}"
    year = f"year={date.year}"
    month = f"month={date.month}"
    day = f"D{date.day}.parquet"
    subdirectory = SUBDIRECTORIES["usgs"]
    return root / subdirectory / state / year / month / day

def generate_usgs_url(
        date: pd.Timestamp,
        state_code: str
) -> str:
    """Returns download URL with parameters."""
    start = date.floor(freq="1d").strftime("%Y-%m-%dT%H:%MZ")
    end = (date.floor(freq="1d") + pd.Timedelta(hours=23, minutes=59)).strftime("%Y-%m-%dT%H:%MZ")
    return NWIS_BASE_URL + f"&stateCd={state_code}&startDT={start}&endDT={end}"

@dataclass
class JSONJob:
    """
    Parameters to keep track of input files and output files for processing
    USGS data.
    """
    ifile: Path
    ofile: Path

def process_json(job: JSONJob) -> None:
    """Process a JSON file."""
    # Load raw data
    with job.ifile.open("r") as fi:
        data = json.loads(fi.read())

    dfs = []
    for site in data["value"]["timeSeries"]:
        usgs_site_code = site["sourceInfo"]["siteCode"][0]["value"]
        for idx, series in enumerate(site["values"]):
            for value in series["value"]:
                # Update series
                value["usgs_site_code"] = usgs_site_code
                value["series"] = idx
                value["qualifiers"] = str(value["qualifiers"])
                dfs.append(value)
    pl.from_dicts(dfs).with_columns(
        pl.col("value").cast(pl.Float32),
        pl.col("usgs_site_code").cast(pl.String),
        pl.col("series").cast(pl.Int32),
        pl.col("qualifiers").cast(pl.String),
        pl.col("dateTime").str.to_datetime("%Y-%m-%dT%H:%M:%S%.3f%:z",
            time_unit="ms").dt.replace_time_zone(None)
    ).rename({
        "value": "observed_cfs",
        "dateTime": "value_time"
    }).drop_nulls("usgs_site_code").write_parquet(job.ofile)

def download_usgs(
        start: pd.Timestamp,
        end: pd.Timestamp,
        root: Path,
        retries: int = 3
    ):
    """
    Download and process NWM output.

    Parameters
    ----------
    start: pandas.Timestamp
        First reference date to retrieve and process.
    end: pandas.Timestamp
        Last refrence date to retrieve and process.
    routelink: polars.DataFrame
        Crosswalk from NWM channel feature IDs to USGS site codes.
    root: pathlib.Path
        Root data directory.
    retries: int, optional, default 3
        Number of times to retry NetCDF file downloads.
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # List of dates to retrieve
    reference_dates = pd.date_range(
        start=start.floor("1d"),
        end=end.ceil("1d"),
        freq="1d"
    )

    # Prepare downloads and output file paths
    urls: list[str] = []
    ofiles: list[Path] = []
    for state in STATE_LIST:
        # Generate file path and check for existence
        state_code = state.abbr.lower()

        # Process each reference day
        for rd in reference_dates:
            # Output file
            ofile = build_usgs_filepath(root, state_code, rd)

            # Skip existing files
            if ofile.exists():
                logger.info("Skipping existing file %s", ofile)
                continue

            logger.info("Preparing to build %s", ofile)
            ofiles.append(ofile)
            ofile.parent.mkdir(exist_ok=True, parents=True)
            urls.append(generate_usgs_url(rd, state_code))

    # Check for files
    if len(ofiles) == 0:
        logger.info("No files to download")
        return

    # Temporary directory
    odir = root / "temp"
    odir.mkdir(exist_ok=True)
    logger.info("Downloading to %s", odir)

    # Download and process
    with TemporaryDirectory(
        prefix="usgs_",
        dir=odir
        ) as td:
        for ofile, url in zip(ofiles, urls):
            # Download file
            json_file = Path(td) / str(ofile).replace(
                "/", "_").replace("=", "_").replace("parquet", "json")

            logger.info("Downloading %s", json_file)
            download_files(
                (url, json_file),
                limit=1,
                timeout=3600,
                headers={"Accept-Encoding": "gzip"},
                file_validator=json_validator,
                retries=retries
            )

            # Process
            logger.info("Building %s", ofile)
            process_json(JSONJob(json_file, ofile))

def scan_usgs_no_cache(root: Path) -> pl.LazyFrame:
    """
    Return polars.LazyFrame of USGS observations.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.

    Returns
    -------
    polars.LazyFrame
    """
    subdirectory = SUBDIRECTORIES["usgs"]
    return pl.scan_parquet(
        root / f"{subdirectory}/",
        hive_schema={
            "state_code": pl.Enum([s.abbr.lower() for s in STATE_LIST]),
            "year": pl.Int32,
            "month": pl.Int32
        }
    )

@functools.lru_cache(LRU_CACHE_SIZES["usgs"])
def scan_usgs_cache(root: Path) -> pl.LazyFrame:
    """
    Return polars.LazyFrame of USGS observations. Cache result.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.

    Returns
    -------
    polars.LazyFrame
    """
    return scan_usgs_no_cache(root)

def scan_usgs(root: Path, cache: bool = False) -> pl.LazyFrame:
    """
    Return polars.LazyFrame of USGS observations.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    cache: bool, optional, default False
        Whether to cache the resulting LazyFrame for subsequent calls.

    Returns
    -------
    polars.LazyFrame
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    logger.info("Scanning observations")
    if cache:
        return scan_usgs_cache(root)
    return scan_usgs_no_cache(root)

def load_usgs_no_cache(
    root: Path,
    state_code: str,
    year: int,
    month: int
    ) -> pl.DataFrame:
    """
    Return polars.DataFrame of USGS observations.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    state_code: str
        Two character lower case state abbreviation.
    year: int
        Integer year.
    month: int
        Integer month.
    
    Returns
    -------
    polars.DataFrame
    """
    return scan_usgs(root).filter(
        pl.col("state_code") == state_code,
        pl.col("year") == year,
        pl.col("month") == month
    ).select(
        ["usgs_site_code", "value_time", "observed_cfs"]
    ).collect().unique(
        ["usgs_site_code", "value_time"], keep="first"
    ).sort(
        ["usgs_site_code", "value_time"]
    )

@functools.lru_cache(LRU_CACHE_SIZES["usgs"])
def load_usgs_cache(
    root: Path,
    state_code: str,
    year: int,
    month: int
    ) -> pl.DataFrame:
    """
    Return polars.DataFrame of USGS observations. Cache DataFrame.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    state_code: str
        Two character lower case state abbreviation.
    year: int
        Integer year.
    month: int
        Integer month.
    
    Returns
    -------
    polars.DataFrame
    """
    return scan_usgs(root, cache=True).filter(
        pl.col("state_code") == state_code,
        pl.col("year") == year,
        pl.col("month") == month
    ).select(
        ["usgs_site_code", "value_time", "observed_cfs"]
    ).collect().unique(
        ["usgs_site_code", "value_time"], keep="first"
    ).sort(
        ["usgs_site_code", "value_time"]
    )

def load_usgs(
    root: Path,
    state_code: str,
    year: int,
    month: int,
    cache: bool = False
    ) -> pl.DataFrame:
    """
    Return polars.DataFrame of USGS observations.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    state_code: str
        Two character lower case state abbreviation.
    year: int
        Integer year.
    month: int
        Integer month.
    cache: bool, optional, default False
        Whether to cache the resulting DataFrame for subsequent calls.
    
    Returns
    -------
    polars.DataFrame
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    logger.info("Retrieving observations")
    if cache:
        return load_usgs_cache(root, state_code, year, month)
    return load_usgs_no_cache(root, state_code, year, month)

def usgs_site_generator(
    root: Path,
    usgs_site_code: str,
    start_time: pd.Timestamp,
    end_time: pd.Timestamp,
    cache: bool = False
    ) -> Generator[pl.DataFrame]:
    """
    Iteratively return polars.DataFrames of USGS observations.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    usgs_site_code: str
        USGS site code.
    start_time: pandas.Timestamp
        Earliest datetime to retrieve.
    end_time: pandas.Timestamp
        Latest datetime to retrieve.
    cache: bool, optional, default False
        Whether to cache the underlying DataFrames for subsequent calls.
    
    Returns
    -------
    polars.DataFrame
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # Find state code
    logger.info("Looking up state code for %s", usgs_site_code)
    state_code = lookup_site_state_code(root, usgs_site_code)

    # Pad retrieval
    padded_start_time = start_time - pd.Timedelta("31D")
    padded_end_time = end_time + pd.Timedelta("31D")

    # Load data
    for m in pd.date_range(padded_start_time, padded_end_time, freq="1ME"):
        # Get month of data
        if cache:
            df = load_usgs_cache(root, state_code, m.year, m.month)
        else:
            df = load_usgs_no_cache(root, state_code, m.year, m.month)

        # Filter to relevant site, add to rest
        yield df.filter(
            pl.col("usgs_site_code") == usgs_site_code,
            pl.col("value_time") >= start_time,
            pl.col("value_time") <= end_time
            ).unique("value_time", keep="first").sort("value_time")

def load_usgs_site(
    root: Path,
    usgs_site_code: str,
    start_time: pd.Timestamp,
    end_time: pd.Timestamp,
    cache: bool = False
    ) -> pl.DataFrame:
    """
    Return polars.DataFrame of USGS observations.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    usgs_site_code: str
        USGS site code.
    start_time: pandas.Timestamp
        Earliest datetime to retrieve.
    end_time: pandas.Timestamp
        Latest datetime to retrieve.
    cache: bool, optional, default False
        Whether to cache the underlying DataFrames for subsequent calls.
    
    Returns
    -------
    polars.DataFrame
    """
    # Collect dataframes
    dataframes = []
    for df in usgs_site_generator(
        root=root,
        usgs_site_code=usgs_site_code,
        start_time=start_time,
        end_time=end_time,
        cache=cache
    ):
        dataframes.append(df)
    return pl.concat(dataframes)



"""Basic functions to build and launch dashboards."""
from pathlib import Path

import panel as pn

from nwm_explorer.dashboard import Dashboard
from nwm_explorer.configuration import load_configuration, Configuration

def generate_dashboard(
        configuration: Configuration
        ) -> Dashboard:
    """Instantiate and return a dashboard."""
    return Dashboard(configuration)

def generate_dashboard_closure(
        configuration: Configuration
        ) -> Dashboard:
    """Build and return a closure function that generates new dashboards."""
    def closure():
        return generate_dashboard(configuration)
    return closure

def serve_dashboards(
        configuration_file: Path
        ) -> None:
    """Serve dashboards."""
    # Load configuration
    configuration = load_configuration(configuration_file)

    # Serve
    endpoints = {
        configuration.endpoint: generate_dashboard_closure(configuration)
    }
    pn.serve(endpoints)



"""
Application-wide configuration.
"""
from pathlib import Path
import inspect
from datetime import datetime

from pydantic import BaseModel

from nwm_explorer.logger import get_logger

class Evaluation(BaseModel):
    """
    Model for evaluation parameters.

    Attributes
    ----------
    label: str
        Machine-friendly label used to generate parquet store.
    start_time: pandas.Timestamp
        First reference time.
    end_time: pandas.Timestamp
        Last reference time.
    """
    label: str
    start_time: datetime
    end_time: datetime

class MapLayer(BaseModel):
    """
    Model for additional map layers.

    Attributes
    ----------
    name: str
        Name to display.
    path: pathlib.Path
        Path to GeoParquet file.
    columns: list[str], optional
        List of column values to display on hover.
    """
    name: str
    path: Path
    columns: list[str] | None = None

class Configuration(BaseModel):
    """
    Application configuration options.

    Attributes
    ----------
    title: str
        Application name.
    endpoint: str
        Endpoint for service.
    root: pathlib.Path
        Root data directory.
    usgs_api_key: str
        USGS API key.
    map_layers: list[MapLayer]
        List of additional map layers to show on map.
    evaluations: list[Evaluation]
        List of evaluations to run.
    processes: int, optional, default 1
        Number of parallel processes to use for computation.
    sites_per_chunk: int, optional, default 1
        Maximum number of sites to load into memory at once.
    retries: int, optional, default 3
        Number of times to retry downloads.
    """
    title: str
    endpoint: str
    root: Path
    usgs_api_key: str
    map_layers: list[MapLayer]
    evaluations: list[Evaluation]
    processes: int = 1
    sites_per_chunk: int = 1
    retries: int = 3

def load_configuration(configuration_file: Path) -> Configuration:
    """
    Returns Configuration object after validating file.

    Parameters
    ----------
    configuration_file: pathlib.Path
        Path to configuration JSON file.

    Returns
    -------
    Configuration
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # Load configuration
    logger.info("Reading %s", configuration_file)
    with configuration_file.open("r") as fo:
        return Configuration.model_validate_json(fo.read())



"""Constants used throughout the application."""
from pathlib import Path
from enum import StrEnum
from dataclasses import dataclass
from typing import TypedDict, Callable

import polars as pl
import us
from plotly.basedatatypes import BaseTraceType
import plotly.graph_objects as go
import numpy as np
import numpy.typing as npt
import pandas as pd

URLBuilder = Callable[[pd.Timestamp], list[str]]
"""A function that takes a pandas.Timestamp and returns a list of URLs."""

class ModelDomainDisplay(StrEnum):
    """Symbols for model domain display."""
    CONUS = "[CONUS] "
    ALASKA = "[Alaska] "
    HAWAII = "[Hawaii] "
    PUERTO_RICO = "[Puerto Rico] "

class ModelDomain(StrEnum):
    """Model domain."""
    ALASKA = "alaska"
    HAWAII = "hawaii"
    CONUS = "conus"
    PUERTO_RICO = "puertorico"

class ModelConfiguration(StrEnum):
    """Model configurations."""
    ANALYSIS_ASSIM_EXTEND_ALASKA_NO_DA = "analysis_assim_extend_alaska_no_da"
    ANALYSIS_ASSIM_EXTEND_NO_DA = "analysis_assim_extend_no_da"
    ANALYSIS_ASSIM_HAWAII_NO_DA = "analysis_assim_hawaii_no_da"
    ANALYSIS_ASSIM_PUERTO_RICO_NO_DA = "analysis_assim_puertorico_no_da"
    MEDIUM_RANGE_MEM_1 = "medium_range_mem1"
    MEDIUM_RANGE_BLEND = "medium_range_blend"
    MEDIUM_RANGE_NO_DA = "medium_range_no_da"
    MEDIUM_RANGE_ALASKA_MEM_1 = "medium_range_alaska_mem1"
    MEDIUM_RANGE_BLEND_ALASKA = "medium_range_blend_alaska"
    MEDIUM_RANGE_ALASKA_NO_DA = "medium_range_alaska_no_da"
    SHORT_RANGE = "short_range"
    SHORT_RANGE_ALASKA = "short_range_alaska"
    SHORT_RANGE_HAWAII = "short_range_hawaii"
    SHORT_RANGE_HAWAII_NO_DA = "short_range_hawaii_no_da"
    SHORT_RANGE_PUERTO_RICO = "short_range_puertorico"
    SHORT_RANGE_PUERTO_RICO_NO_DA = "short_range_puertorico_no_da"

class Metric(StrEnum):
    """Symbols for common metrics."""
    NASH_SUTCLIFFE_EFFICIENCY = "nash_sutcliffe_efficiency"
    RELATIVE_MEAN_BIAS = "relative_mean_bias"
    PEARSON_CORRELATION_COEFFICIENT = "pearson_correlation_coefficient"
    RELATIVE_MEAN = "relative_mean"
    RELATIVE_MEDIAN = "relative_median"
    RELATIVE_MINIMUM = "relative_minimum"
    RELATIVE_MAXIMUM = "relative_maximum"
    RELATIVE_STANDARD_DEVIATION = "relative_standard_deviation"
    KLING_GUPTA_EFFICIENCY = "kling_gupta_efficiency"

class SiteTypeSlug(StrEnum):
    """Machine-friendly site types."""
    STREAM = "stream"
    CANAL = "canal"
    DITCH = "ditch"
    LAKE = "lake"
    TIDAL = "tidal"
    SPRING = "spring"
    DIVERSION = "diversion"
    ESTUARY = "estuary"
    TUNNEL = "tunnel"
    FIELD = "field"
    STORM_SEWER = "storm_sewer"
    COMBINED_SEWER = "combined_sewer"
    OUTFALL = "outfall"

@dataclass
class SiteType:
    """
    Dataclass for storing USGS site type details.

    Attributes
    ----------
    code: str
        Site type code.
    name: str
        Short name.
    long_name: str
        Long name.
    slug: SiteTypeSlug
        Machine-friendly representation.
    """
    code: str
    name: str
    long_name: str
    slug: SiteTypeSlug

@dataclass
class NWMGroupSpecification:
    """
    A dataclass that holds specifications for time-based polars groupby
    operations.
    """
    index_column: str = "value_time"
    group_by_columns: list[str] | None = None
    select_columns: list[str] | None = None
    sort_columns: list[str] | None = None
    window_interval: int = 24
    state_code: str | None = None
    lead_time_hours_max: int = 0

    def __post_init__(self) -> None:
        if self.group_by_columns is None:
            self.group_by_columns = ["nwm_feature_id", "reference_time"]

        if self.select_columns is None:
            self.select_columns = self.group_by_columns + [self.index_column, "predicted_cfs"]

        if self.sort_columns is None:
            self.sort_columns = self.group_by_columns + [self.index_column]

ROUTELINK_URL: str = (
    "https://www.hydroshare.org/resource"
    "/1fe9975004ce4b5097d41939afa14f84/data/contents/RouteLinks.tar.gz"
)
"""URL to RouteLink CSV tarball."""

ROUTELINK_PARQUET: Path = Path("routelink.parquet")
"""Default path to polars-compatible RouteLink parquet file used by application."""

ROUTELINK_FILENAMES: dict[ModelDomain, str] = {
    ModelDomain.ALASKA: "RouteLink_AK.csv",
    ModelDomain.CONUS: "RouteLink_CONUS.csv",
    ModelDomain.HAWAII: "RouteLink_HI.csv",
    ModelDomain.PUERTO_RICO: "RouteLink_PRVI.csv"
}
"""Mapping from domains to routelink files names."""

GOOGLE_CLOUD_BUCKET_URL: str = "https://storage.googleapis.com/national-water-model/"
"""National Water Model Google Cloud Storage bucket."""

DOMAIN_LOOKUP: dict[ModelConfiguration, ModelDomainDisplay] = {
    ModelConfiguration.ANALYSIS_ASSIM_EXTEND_NO_DA: ModelDomainDisplay.CONUS,
    ModelConfiguration.MEDIUM_RANGE_MEM_1: ModelDomainDisplay.CONUS,
    ModelConfiguration.MEDIUM_RANGE_BLEND: ModelDomainDisplay.CONUS,
    ModelConfiguration.MEDIUM_RANGE_NO_DA: ModelDomainDisplay.CONUS,
    ModelConfiguration.SHORT_RANGE: ModelDomainDisplay.CONUS,
    ModelConfiguration.ANALYSIS_ASSIM_EXTEND_ALASKA_NO_DA: ModelDomainDisplay.ALASKA,
    ModelConfiguration.MEDIUM_RANGE_ALASKA_MEM_1: ModelDomainDisplay.ALASKA,
    ModelConfiguration.MEDIUM_RANGE_BLEND_ALASKA: ModelDomainDisplay.ALASKA,
    ModelConfiguration.MEDIUM_RANGE_ALASKA_NO_DA: ModelDomainDisplay.ALASKA,
    ModelConfiguration.SHORT_RANGE_ALASKA: ModelDomainDisplay.ALASKA,
    ModelConfiguration.ANALYSIS_ASSIM_HAWAII_NO_DA: ModelDomainDisplay.HAWAII,
    ModelConfiguration.SHORT_RANGE_HAWAII: ModelDomainDisplay.HAWAII,
    ModelConfiguration.SHORT_RANGE_HAWAII_NO_DA: ModelDomainDisplay.HAWAII,
    ModelConfiguration.ANALYSIS_ASSIM_PUERTO_RICO_NO_DA: ModelDomainDisplay.PUERTO_RICO,
    ModelConfiguration.SHORT_RANGE_PUERTO_RICO: ModelDomainDisplay.PUERTO_RICO,
    ModelConfiguration.SHORT_RANGE_PUERTO_RICO_NO_DA: ModelDomainDisplay.PUERTO_RICO
}
"""Mapping from ModelConfiguration to ModelDomainDisplay."""

CONFIGURATION_LOOKUP: dict[ModelConfiguration, str] = {
    ModelConfiguration.ANALYSIS_ASSIM_EXTEND_NO_DA: "Extended Analysis & Assimilation"
                                                    " (MRMS/Stage IV, No-DA)",
    ModelConfiguration.MEDIUM_RANGE_MEM_1: "Medium Range Deterministic (GFS)",
    ModelConfiguration.MEDIUM_RANGE_BLEND: "Medium Range Deterministic (NBM)",
    ModelConfiguration.MEDIUM_RANGE_NO_DA: "Medium Range Deterministic (GFS, No-DA)",
    ModelConfiguration.SHORT_RANGE: "Short Range (HRRR)",
    ModelConfiguration.ANALYSIS_ASSIM_EXTEND_ALASKA_NO_DA: "Extended Analysis & Assimilation"
                                                           " (MRMS/Stage IV, No-DA)",
    ModelConfiguration.MEDIUM_RANGE_ALASKA_MEM_1: "Medium Range Deterministic (GFS)",
    ModelConfiguration.MEDIUM_RANGE_BLEND_ALASKA: "Medium Range Deterministic (NBM)",
    ModelConfiguration.MEDIUM_RANGE_ALASKA_NO_DA: "Medium Range Deterministic (GFS, No-DA)",
    ModelConfiguration.SHORT_RANGE_ALASKA: "Short Range (HRRR)",
    ModelConfiguration.ANALYSIS_ASSIM_HAWAII_NO_DA: "Analysis & Assimilation (MRMS, No-DA)",
    ModelConfiguration.SHORT_RANGE_HAWAII: "Short Range (WRF-ARW)",
    ModelConfiguration.SHORT_RANGE_HAWAII_NO_DA: "Short Range (WRF-ARW, No-DA)",
    ModelConfiguration.ANALYSIS_ASSIM_PUERTO_RICO_NO_DA: "Analysis & Assimilation (MRMS, No-DA)",
    ModelConfiguration.SHORT_RANGE_PUERTO_RICO: "Short Range (WRF-ARW)",
    ModelConfiguration.SHORT_RANGE_PUERTO_RICO_NO_DA: "Short Range (WRF-ARW, No-DA)"
}
"""Mapping from ModelConfiguration to pretty strings."""

GROUP_SPECIFICATIONS: dict[ModelConfiguration, NWMGroupSpecification] = {
    ModelConfiguration.ANALYSIS_ASSIM_EXTEND_ALASKA_NO_DA: NWMGroupSpecification(
        group_by_columns=["nwm_feature_id"],
        state_code="ak",
        select_columns=["nwm_feature_id", "reference_time", "value_time", "predicted_cfs"]
    ),
    ModelConfiguration.MEDIUM_RANGE_ALASKA_MEM_1: NWMGroupSpecification(
        state_code="ak",
        lead_time_hours_max=240
    ),
    ModelConfiguration.MEDIUM_RANGE_BLEND_ALASKA: NWMGroupSpecification(
        state_code="ak",
        lead_time_hours_max=240
    ),
    ModelConfiguration.MEDIUM_RANGE_ALASKA_NO_DA: NWMGroupSpecification(
        state_code="ak",
        lead_time_hours_max=240
    ),
    ModelConfiguration.SHORT_RANGE_ALASKA: NWMGroupSpecification(
        window_interval=5,
        state_code="ak",
        lead_time_hours_max=45
    ),
    ModelConfiguration.ANALYSIS_ASSIM_HAWAII_NO_DA: NWMGroupSpecification(
        group_by_columns=["nwm_feature_id"],
        state_code="hi",
        select_columns=["nwm_feature_id", "reference_time", "value_time", "predicted_cfs"]
    ),
    ModelConfiguration.SHORT_RANGE_HAWAII: NWMGroupSpecification(
        window_interval=6,
        state_code="hi",
        lead_time_hours_max=48
    ),
    ModelConfiguration.SHORT_RANGE_HAWAII_NO_DA: NWMGroupSpecification(
        window_interval=6,
        state_code="hi",
        lead_time_hours_max=48
    ),
    ModelConfiguration.ANALYSIS_ASSIM_PUERTO_RICO_NO_DA: NWMGroupSpecification(
        group_by_columns=["nwm_feature_id"],
        state_code="pr",
        select_columns=["nwm_feature_id", "reference_time", "value_time", "predicted_cfs"]
    ),
    ModelConfiguration.SHORT_RANGE_PUERTO_RICO: NWMGroupSpecification(
        window_interval=6,
        state_code="pr",
        lead_time_hours_max=48
    ),
    ModelConfiguration.SHORT_RANGE_PUERTO_RICO_NO_DA: NWMGroupSpecification(
        window_interval=6,
        state_code="pr",
        lead_time_hours_max=48
    ),
    ModelConfiguration.ANALYSIS_ASSIM_EXTEND_NO_DA: NWMGroupSpecification(
        group_by_columns=["nwm_feature_id"],
        select_columns=["nwm_feature_id", "reference_time", "value_time", "predicted_cfs"]
    ),
    ModelConfiguration.MEDIUM_RANGE_MEM_1: NWMGroupSpecification(lead_time_hours_max=240),
    ModelConfiguration.MEDIUM_RANGE_BLEND: NWMGroupSpecification(lead_time_hours_max=240),
    ModelConfiguration.MEDIUM_RANGE_NO_DA: NWMGroupSpecification(lead_time_hours_max=240),
    ModelConfiguration.SHORT_RANGE: NWMGroupSpecification(
        window_interval=6,
        lead_time_hours_max=18
    )
}
"""Mapping from ModelConfiguration to group-by specifications."""

METRIC_LOOKUP: dict[str, Metric] = {
    "Kling-Gupta efficiency": Metric.KLING_GUPTA_EFFICIENCY,
    "Nash-Sutcliffe efficiency": Metric.NASH_SUTCLIFFE_EFFICIENCY,
    "Relative mean": Metric.RELATIVE_MEAN,
    "Relative standard deviation": Metric.RELATIVE_STANDARD_DEVIATION,
    "Pearson correlation coefficient": Metric.PEARSON_CORRELATION_COEFFICIENT,
    "Relative mean bias": Metric.RELATIVE_MEAN_BIAS,
    "Relative median": Metric.RELATIVE_MEDIAN,
    "Relative minimum": Metric.RELATIVE_MINIMUM,
    "Relative maximum": Metric.RELATIVE_MAXIMUM
}
"""Mapping from pretty strings to evaluation Metric."""

RANK_LOOKUP: dict[str, str] = {
    "Median": "median",
    "Minimum": "min",
    "Maximum": "max"
}
"""Mapping from pretty strings to column label components. 'Rank' refers to the
aggregation method used to resample streamflow.
"""

DEFAULT_ZOOM: dict[ModelDomainDisplay, int] = {
    ModelDomainDisplay.ALASKA: 5,
    ModelDomainDisplay.CONUS: 3,
    ModelDomainDisplay.HAWAII: 6,
    ModelDomainDisplay.PUERTO_RICO: 8
}
"""Default map zoom for each domain."""

DEFAULT_CENTER: dict[ModelDomainDisplay, dict[str, float]] = {
    ModelDomainDisplay.ALASKA: {"lat": 60.84683, "lon": -149.05659},
    ModelDomainDisplay.CONUS: {"lat": 38.83348, "lon": -93.97612},
    ModelDomainDisplay.HAWAII: {"lat": 21.24988, "lon": -157.59606},
    ModelDomainDisplay.PUERTO_RICO: {"lat": 18.21807, "lon": -66.32802}
}
"""Default map center for each domain."""

METRIC_PLOTTING_LIMITS: dict[Metric, tuple[float, float]] = {
    Metric.RELATIVE_MEAN_BIAS: (-1.0, 1.0),
    Metric.PEARSON_CORRELATION_COEFFICIENT: (-1.0, 1.0),
    Metric.NASH_SUTCLIFFE_EFFICIENCY: (-1.0, 1.0),
    Metric.RELATIVE_MEAN: (0.0, 2.0),
    Metric.RELATIVE_STANDARD_DEVIATION: (0.0, 2.0),
    Metric.RELATIVE_MEDIAN: (0.0, 2.0),
    Metric.RELATIVE_MINIMUM: (0.0, 2.0),
    Metric.RELATIVE_MAXIMUM: (0.0, 2.0),
    Metric.KLING_GUPTA_EFFICIENCY: (-1.0, 1.0)
}
"""Mapping from Metrics to plotting limits (cmin, cmax)."""

METRIC_SIGNIFICANCE_THRESHOLD: dict[Metric, float] = {
    Metric.RELATIVE_MEAN_BIAS: 0.0,
    Metric.PEARSON_CORRELATION_COEFFICIENT: 0.0,
    Metric.NASH_SUTCLIFFE_EFFICIENCY: 0.0,
    Metric.RELATIVE_MEAN: 1.0,
    Metric.RELATIVE_STANDARD_DEVIATION: 1.0,
    Metric.RELATIVE_MEDIAN: 1.0,
    Metric.RELATIVE_MINIMUM: 1.0,
    Metric.RELATIVE_MAXIMUM: 1.0,
    Metric.KLING_GUPTA_EFFICIENCY: 0.0
}
"""
Mapping from Metrics to a value that if it falls outside the 95% confidence interval,
indicates 'statistical significance.'
"""

LRU_CACHE_SIZES: dict[str, int] = {
    "evaluations": 10,
    "nwm": 9,
    "usgs": 25,
    "pairs": 60
}
"""Maximum size of functools.lru_cache."""

SUBDIRECTORIES: dict[str, str] = {
    "evaluations": "evaluations",
    "nwm": "nwm",
    "usgs": "usgs",
    "site_table": "site_table",
    "pairs": "pairs"
}
"""Subdirectories that indicates root of parquet stores."""

NWIS_BASE_URL: str = (
    "https://waterservices.usgs.gov/nwis/iv/"
    "?format=json&siteStatus=all&parameterCd=00060"
)
"""NWIS IV API returning json and all site statuses."""

MONITORING_LOCATION_BASE_URL: str = (
    "https://api.waterdata.usgs.gov/ogcapi/v0/collections/monitoring-locations"
    "/items?f=json&lang=en-US&limit=10000&skipGeometry=false&offset=0"
    "&agency_code=USGS&site_type_code="
)
"""USGS monitoring location API returning geojson."""

SITE_TYPES: list[SiteType] = [
    SiteType("ST", "Stream", "Stream", SiteTypeSlug.STREAM),
    SiteType("ST-CA", "Canal", "Canal", SiteTypeSlug.CANAL),
    SiteType("ST-DCH", "Ditch", "Ditch", SiteTypeSlug.DITCH),
    SiteType("ST-TS", "Tidal SW", "Tidal stream", SiteTypeSlug.TIDAL),
    SiteType("LK", "Lake", "Lake, Reservoir, Impoundment", SiteTypeSlug.LAKE),
    SiteType("SP", "Spring", "Spring", SiteTypeSlug.SPRING),
    SiteType("FA-DV", "Diversion", "Diversion", SiteTypeSlug.DIVERSION),
    SiteType("ES", "Estuary", "Estuary", SiteTypeSlug.ESTUARY),
    SiteType("SB-TSM", "Tunl/mine", "Tunnel, shaft, or mine", SiteTypeSlug.TUNNEL),
    SiteType("FA-FON", "Agric area", "Field, Pasture, Orchard, or Nursery", SiteTypeSlug.FIELD),
    SiteType("FA-STS", "Sewer-strm", "Storm sewer", SiteTypeSlug.STORM_SEWER),
    SiteType("FA-CS", "Sewer-comb", "Combined sewer", SiteTypeSlug.COMBINED_SEWER),
    SiteType("FA-OF", "Outfall", "Outfall", SiteTypeSlug.OUTFALL)
]
"""List of USGS site types to retrieve for master site table."""

STATE_LIST: list[us.states.State] = us.states.STATES + [us.states.PR, us.states.DC]
"""List of US states."""

SITE_SCHEMA: pl.Schema = pl.Schema({
    "id": pl.String,
    "vertical_datum": pl.String,
    "original_horizontal_datum_name": pl.String,
    "well_constructed_depth": pl.String,
    "country_name": pl.String,
    "vertical_datum_name": pl.String,
    "drainage_area": pl.Float64,
    "hole_constructed_depth": pl.String,
    "minor_civil_division_code": pl.String,
    "hydrologic_unit_code": pl.String,
    "horizontal_positional_accuracy_code": pl.String,
    "contributing_drainage_area": pl.Float64,
    "depth_source_code": pl.String,
    "agency_name": pl.String,
    "basin_code": pl.String,
    "horizontal_positional_accuracy": pl.String,
    "time_zone_abbreviation": pl.String,
    "altitude": pl.Float64,
    "monitoring_location_name": pl.String,
    "district_code": pl.String,
    "state_code": pl.String,
    "site_type": pl.String,
    "horizontal_position_method_code": pl.String,
    "uses_daylight_savings": pl.String,
    "agency_code": pl.String,
    "country_code": pl.String,
    "county_code": pl.String,
    "altitude_accuracy": pl.Float64,
    "construction_date": pl.String,
    "aquifer_code": pl.String,
    "monitoring_location_number": pl.String,
    "state_name": pl.String,
    "site_type_code": pl.String,
    "altitude_method_code": pl.String,
    "horizontal_position_method_name": pl.String,
    "national_aquifer_code": pl.String,
    "county_name": pl.String,
    "altitude_method_name": pl.String,
    "original_horizontal_datum": pl.String,
    "aquifer_type_code": pl.String,
    "longitude": pl.Float64,
    "latitude": pl.Float64
})
"""Schema for USGS site data."""

class PlotlyFigure(TypedDict):
    """
    Specifies plotly figure dict for use with panel.

    Attributes
    ----
    data: list[plotly.basedatatypes.BaseTraceType]
        List of plotly traces.
    layout: plotly.graph_objects.Layout
        Plotly layout.
    """
    data: list[BaseTraceType]
    layout: go.Layout

DEFAULT_ZOOM: dict[ModelDomainDisplay, int] = {
    ModelDomainDisplay.ALASKA: 5,
    ModelDomainDisplay.CONUS: 3,
    ModelDomainDisplay.HAWAII: 6,
    ModelDomainDisplay.PUERTO_RICO: 8
}
"""Default map zoom for each domain."""

DEFAULT_CENTER: dict[ModelDomainDisplay, dict[str, float]] = {
    ModelDomainDisplay.ALASKA: {"lat": 60.84683, "lon": -149.05659},
    ModelDomainDisplay.CONUS: {"lat": 38.83348, "lon": -93.97612},
    ModelDomainDisplay.HAWAII: {"lat": 21.24988, "lon": -157.59606},
    ModelDomainDisplay.PUERTO_RICO: {"lat": 18.21807, "lon": -66.32802}
}
"""Default map center for each domain."""

METRIC_PLOTTING_LIMITS: dict[Metric, tuple[float, float]] = {
    Metric.RELATIVE_MEAN_BIAS: (-1.0, 1.0),
    Metric.PEARSON_CORRELATION_COEFFICIENT: (-1.0, 1.0),
    Metric.NASH_SUTCLIFFE_EFFICIENCY: (-1.0, 1.0),
    Metric.RELATIVE_MEAN: (0.0, 2.0),
    Metric.RELATIVE_STANDARD_DEVIATION: (0.0, 2.0),
    Metric.RELATIVE_MEDIAN: (0.0, 2.0),
    Metric.RELATIVE_MINIMUM: (0.0, 2.0),
    Metric.RELATIVE_MAXIMUM: (0.0, 2.0),
    Metric.KLING_GUPTA_EFFICIENCY: (-1.0, 1.0)
}
"""Mapping from Metrics to plotting limits (cmin, cmax)."""

CONFIGURATION_LINE_TYPE: dict[ModelConfiguration, str] = {
    ModelConfiguration.ANALYSIS_ASSIM_PUERTO_RICO_NO_DA: "markers"
}
"""Mapping from Model configuration to line type for plotting."""

class AxisType(StrEnum):
    """Plotly axis type."""
    NONE = "-"
    LINEAR = "linear"
    LOG = "log"
    DATE = "date"
    CATEGORY = "category"
    MULTICATEGORY = "multicategory"

MetricFunction = Callable[[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64]
    ], None]
"""Type hint for Numba functions that generate metrics."""

SITE_COLUMN_MAPPING: dict[str, str] = {
    "monitoring_location_name": "Name",
    "monitoring_location_number": "Site code",
    "hydrologic_unit_code": "HUC",
    "site_type": "Site type",
    "drainage_area": "Drainage area (sq.mi.)",
    "contributing_drainage_area": "Contrib. drain. area (sq.mi.)"
}
"""Mapping from site data column to pretty string."""

class MeasurementUnits(StrEnum):
    """Streamflow measurement units."""
    CUBIC_FEET_PER_SECOND = "CFS"
    CUBIC_METERS_PER_SECOND = "CMS"
    CUBIC_FEET_PER_SECOND_PER_SQUARE_MILE = "CFS/sq.mi."
    INCHES_PER_HOUR = "inch/h"
    CUMULATIVE_INCHES_PER_HOUR = "inch/h, accum."

COLUMN_DESCRIPTIONS: dict[str, str] = {
    "label": "Evaluation label",
    "configuration": "National Water Model configuration",
    "nwm_feature_id": "National Water Model channel feature identifier",
    "lead_time_hours_min": "Minimum lead time hours, valid time minus reference time",
    "predicted_cfs_min": "Minimum predicted streamflow in cubic feet per second",
    "predicted_cfs_median": "Median predicted streamflow in cubic feet per second",
    "predicted_cfs_max": "Maximum predicted streamflow in cubic feet per second",
    "predicted_value_time_min": "Earliest prediction valid time",
    "predicted_value_time_max": "Latest prediction valid time",
    "reference_time_min": "Earliest prediction reference time, also called issue time",
    "reference_time_max": "Latest prediction reference time, also called issue time",
    "usgs_site_code": "USGS site code",
    "observed_cfs_min": "Minimum observed streamflow in cubic feet per second",
    "observed_cfs_median": "Median observed streamflow in cubic feet per second",
    "observed_cfs_max": "Maximum observed streamflow in cubic feet per second",
    "observed_value_time_min": "Earliest observation valid time",
    "observed_value_time_max": "Latest observation valid time",
    "sample_size": "Number of prediction-observation pairs",
    "nash_sutcliffe_efficiency_min_point": "Nash-Sutcliffe model efficiency of minimum predicted streamflow, point estimate",
    "nash_sutcliffe_efficiency_min_lower": "Nash-Sutcliffe model efficiency of minimum predicted streamflow, lower 95% confidence estimate",
    "nash_sutcliffe_efficiency_min_upper": "Nash-Sutcliffe model efficiency of minimum predicted streamflow, upper 95% confidence estimate",
    "relative_mean_bias_min_point": "Relative mean bias/error of minimum predicted streamflow, point estimate",
    "relative_mean_bias_min_lower": "Relative mean bias/error of minimum predicted streamflow, lower 95% confidence estimate",
    "relative_mean_bias_min_upper": "Relative mean bias/error of minimum predicted streamflow, upper 95% confidence estimate",
    "pearson_correlation_coefficient_min_point": "Pearson correlation coefficient of minimum predicted streamflow, point estimate",
    "pearson_correlation_coefficient_min_lower": "Pearson correlation coefficient of minimum predicted streamflow, lower 95% confidence estimate",
    "pearson_correlation_coefficient_min_upper": "Pearson correlation coefficient of minimum predicted streamflow, upper 95% confidence estimate",
    "relative_mean_min_point": "Relative mean of minimum predicted streamflow, point estimate",
    "relative_mean_min_lower": "Relative mean of minimum predicted streamflow, lower 95% confidence estimate",
    "relative_mean_min_upper": "Relative mean of minimum predicted streamflow, upper 95% confidence estimate",
    "relative_median_min_point": "Relative median of minimum predicted streamflow, point estimate",
    "relative_median_min_lower": "Relative median of minimum predicted streamflow, lower 95% confidence estimate",
    "relative_median_min_upper": "Relative median of minimum predicted streamflow, upper 95% confidence estimate",
    "relative_minimum_min_point": "Relative minimum of minimum predicted streamflow, point estimate",
    "relative_minimum_min_lower": "Relative minimum of minimum predicted streamflow, lower 95% confidence estimate",
    "relative_minimum_min_upper": "Relative minimum of minimum predicted streamflow, upper 95% confidence estimate",
    "relative_maximum_min_point": "Relative maximum of minimum predicted streamflow, point estimate",
    "relative_maximum_min_lower": "Relative maximum of minimum predicted streamflow, lower 95% confidence estimate",
    "relative_maximum_min_upper": "Relative maximum of minimum predicted streamflow, upper 95% confidence estimate",
    "relative_standard_deviation_min_point": "Relative standard deviation of minimum predicted streamflow, point estimate",
    "relative_standard_deviation_min_lower": "Relative standard deviation of minimum predicted streamflow, lower 95% confidence estimate",
    "relative_standard_deviation_min_upper": "Relative standard deviation of minimum predicted streamflow, upper 95% confidence estimate",
    "kling_gupta_efficiency_min_point": "Kling-Gupta model efficiency of minimum predicted streamflow, point estimate",
    "kling_gupta_efficiency_min_lower": "Kling-Gupta model efficiency of minimum predicted streamflow, lower 95% confidence estimate",
    "kling_gupta_efficiency_min_upper": "Kling-Gupta model efficiency of minimum predicted streamflow, upper 95% confidence estimate",
    "nash_sutcliffe_efficiency_median_point": "Nash-Sutcliffe model efficiency of median predicted streamflow, point estimate",
    "nash_sutcliffe_efficiency_median_lower": "Nash-Sutcliffe model efficiency of median predicted streamflow, lower 95% confidence estimate",
    "nash_sutcliffe_efficiency_median_upper": "Nash-Sutcliffe model efficiency of median predicted streamflow, upper 95% confidence estimate",
    "relative_mean_bias_median_point": "Relative mean bias/error of median predicted streamflow, point estimate",
    "relative_mean_bias_median_lower": "Relative mean bias/error of median predicted streamflow, lower 95% confidence estimate",
    "relative_mean_bias_median_upper": "Relative mean bias/error of median predicted streamflow, upper 95% confidence estimate",
    "pearson_correlation_coefficient_median_point": "Pearson correlation coefficient of median predicted streamflow, point estimate",
    "pearson_correlation_coefficient_median_lower": "Pearson correlation coefficient of median predicted streamflow, lower 95% confidence estimate",
    "pearson_correlation_coefficient_median_upper": "Pearson correlation coefficient of median predicted streamflow, upper 95% confidence estimate",
    "relative_mean_median_point": "Relative mean of median predicted streamflow, point estimate",
    "relative_mean_median_lower": "Relative mean of median predicted streamflow, lower 95% confidence estimate",
    "relative_mean_median_upper": "Relative mean of median predicted streamflow, point estimate",
    "relative_median_median_point": "Relative median of median predicted streamflow, point estimate",
    "relative_median_median_lower": "Relative median of median predicted streamflow, lower 95% confidence estimate",
    "relative_median_median_upper": "Relative median of median predicted streamflow, point estimate",
    "relative_minimum_median_point": "Relative minimum of median predicted streamflow, point estimate",
    "relative_minimum_median_lower": "Relative minimum of median predicted streamflow, lower 95% confidence estimate",
    "relative_minimum_median_upper": "Relative minimum of median predicted streamflow, point estimate",
    "relative_maximum_median_point": "Relative maximum of median predicted streamflow, point estimate",
    "relative_maximum_median_lower": "Relative maximum of median predicted streamflow, lower 95% confidence estimate",
    "relative_maximum_median_upper": "Relative maximum of median predicted streamflow, upper 95% confidence estimate",
    "relative_standard_deviation_median_point": "Relative standard deviation of median predicted streamflow, point estimate",
    "relative_standard_deviation_median_lower": "Relative standard deviation of median predicted streamflow, lower 95% confidence estimate",
    "relative_standard_deviation_median_upper": "Relative standard deviation of median predicted streamflow, upper 95% confidence estimate",
    "kling_gupta_efficiency_median_point": "Kling-Gupta model efficiency of median predicted streamflow, point estimate",
    "kling_gupta_efficiency_median_lower": "Kling-Gupta model efficiency of median predicted streamflow, lower 95% confidence estimate",
    "kling_gupta_efficiency_median_upper": "Kling-Gupta model efficiency of median predicted streamflow, upper 95% confidence estimate",
    "nash_sutcliffe_efficiency_max_point": "Nash-Sutcliffe model efficiency of maximum predicted streamflow, point estimate",
    "nash_sutcliffe_efficiency_max_lower": "Nash-Sutcliffe model efficiency of maximum predicted streamflow, lower 95% confidence estimate",
    "nash_sutcliffe_efficiency_max_upper": "Nash-Sutcliffe model efficiency of maximum predicted streamflow, upper 95% confidence estimate",
    "relative_mean_bias_max_point": "Relative mean bias/error of maximum predicted streamflow, point estimate",
    "relative_mean_bias_max_lower": "Relative mean bias/error of maximum predicted streamflow, lower 95% confidence estimate",
    "relative_mean_bias_max_upper": "Relative mean bias/error of maximum predicted streamflow, upper 95% confidence estimate",
    "pearson_correlation_coefficient_max_point": "Pearson correlation coefficient of maximum predicted streamflow, point estimate",
    "pearson_correlation_coefficient_max_lower": "Pearson correlation coefficient of maximum predicted streamflow, lower 95% confidence estimate",
    "pearson_correlation_coefficient_max_upper": "Pearson correlation coefficient of maximum predicted streamflow, upper 95% confidence estimate",
    "relative_mean_max_point": "Relative mean of maximum predicted streamflow, point estimate",
    "relative_mean_max_lower": "Relative mean of maximum predicted streamflow, lower 95% confidence estimate",
    "relative_mean_max_upper": "Relative mean of maximum predicted streamflow, upper 95% confidence estimate",
    "relative_median_max_point": "Relative median of maximum predicted streamflow, point estimate",
    "relative_median_max_lower": "Relative median of maximum predicted streamflow, lower 95% confidence estimate",
    "relative_median_max_upper": "Relative median of maximum predicted streamflow, upper 95% confidence estimate",
    "relative_minimum_max_point": "Relative minimum of maximum predicted streamflow, point estimate",
    "relative_minimum_max_lower": "Relative minimum of maximum predicted streamflow, lower 95% confidence estimate",
    "relative_minimum_max_upper": "Relative minimum of maximum predicted streamflow, upper 95% confidence estimate",
    "relative_maximum_max_point": "Relative maximum of maximum predicted streamflow, point estimate",
    "relative_maximum_max_lower": "Relative maximum of maximum predicted streamflow, lower 95% confidence estimate",
    "relative_maximum_max_upper": "Relative maximum of maximum predicted streamflow, upper 95% confidence estimate",
    "relative_standard_deviation_max_point": "Relative standard deviation of maximum predicted streamflow, point estimate",
    "relative_standard_deviation_max_lower": "Relative standard deviation of maximum predicted streamflow, lower 95% confidence estimate",
    "relative_standard_deviation_max_upper": "Relative standard deviation of maximum predicted streamflow, upper 95% confidence estimate",
    "kling_gupta_efficiency_max_point": "Kling-Gupta model efficiency of maximum predicted streamflow, point estimate",
    "kling_gupta_efficiency_max_lower": "Kling-Gupta model efficiency of maximum predicted streamflow, lower 95% confidence estimate",
    "kling_gupta_efficiency_max_upper": "Kling-Gupta model efficiency of maximum predicted streamflow, upper 95% confidence estimate"
}
"""Mapping from columns to descriptions."""



"""
Download and process National Water Model output.

Methods
-------
- download_nwm
- scan_nwm
"""
from tempfile import TemporaryDirectory
from typing import Callable, Generator
from dataclasses import dataclass
from pathlib import Path
import inspect
from concurrent.futures import ProcessPoolExecutor
import os
import functools
import warnings

import numpy as np
import pandas as pd
import polars as pl
import xarray as xr

from nwm_explorer.logger import get_logger
from nwm_explorer.downloads import download_files, FileValidationError
from nwm_explorer.routelink import download_routelink
from nwm_explorer.constants import (LRU_CACHE_SIZES, SUBDIRECTORIES, GOOGLE_CLOUD_BUCKET_URL,
    ModelConfiguration, ModelDomain, URLBuilder)

def netcdf_validator(filepath: Path) -> None:
    """
    Validate that given filepath opens and closes without raising.

    Parameters
    ----------
    filepath: Path
        Path to file.
    
    Returns
    -------
    None

    Raises
    ------
    FileValidationError
    """
    try:
        with xr.open_dataset(filepath):
            return
    except Exception as e:
        raise FileValidationError(e) from e

@dataclass
class NetCDFJob:
    """
    Input data for NetCDF processing jobs. Intended for use with National
    Water Model output.

    Attributes
    ----------
    filepaths: list[Path]
        List of filepaths to process.
    variables: list[str]
        Variables to extract from NetCDF Files.
    features: list[int]
        Feature to extract from NetCDF Files.
    """
    filepaths: list[Path]
    variables: list[str]
    features: list[int]

def process_netcdf(
        job: NetCDFJob
    ) -> pd.DataFrame:
    """
    Process a collection of National Water Model NetCDF files and return a
    dataframe.

    Parameters
    ----------
    job: NetCDFJob
        Job object used to track input files, target variables, and features.

    Returns
    -------
    pandas.DataFrame
    """
    dfs = []
    for fp in job.filepaths:
        # Extract data
        with xr.open_dataset(fp, engine="h5netcdf") as ds:
            df = ds[job.variables].sel(feature_id=job.features
                ).to_dataframe().reset_index().dropna()
            if "time" not in df:
                df["time"] = ds.time.values[0]
            if "reference_time" not in df:
                df["reference_time"] = ds.reference_time.values[0]

        # Add to list
        dfs.append(df)

    # Merge
    merged = pd.concat(dfs, ignore_index=True)

    # Update columns
    merged = merged.rename(columns={
        "time": "value_time",
        "feature_id": "nwm_feature_id",
        "streamflow": "predicted_cfs"
        })

    # Downcast and convert to cubic feet per second
    merged["predicted_cfs"] = merged["predicted_cfs"].astype(np.float32) / (0.3048 ** 3.0)
    return merged

def process_netcdf_parallel(
    filepaths: list[Path],
    variables: list[str],
    features: list[int],
    max_processes: int = 1
    ) -> pd.DataFrame:
    """
    Process a collection of National Water Model NetCDF files and return a
    dataframe, in parallel.

    Parameters
    ----------
    filepaths: list[Path]
        List of filepaths to process.
    variables: list[str]
        Variables to extract from NetCDF Files.
    features: list[int]
        Feature to extract from NetCDF Files.
    max_processes: int, optional, default 1
        Maximum number of cores to use simultaneously.

    Returns
    -------
    pandas.DataFrame
    """
    job_files = np.array_split(filepaths, max_processes)
    jobs = [NetCDFJob(j, variables, features) for j in job_files]
    with ProcessPoolExecutor(max_workers=max_processes) as pool:
        return pd.concat(pool.map(process_netcdf, jobs), ignore_index=True)

def generate_reference_dates(
        start: str | pd.Timestamp,
        end: str | pd.Timestamp
) -> list[pd.Timestamp]:
    """
    Return list of pandas.Timestamp from start
    date to end date.

    Parameters
    ----------
    start: str | Timestamp, required
        First date.
    end: str | Timestamp, required
        Last date
    
    Returns
    -------
    list[pd.Timestamp]
    """
    return pd.date_range(
        start=start.tz_localize(None).floor(freq="1d"),
        end=end.tz_localize(None).floor(freq="1d"),
        freq="1d"
    ).to_list()

def build_gcs_public_urls(
        reference_date: pd.Timestamp,
        configuration: str,
        prefixes: list[str],
        file_type: str,
        suffix: str,
        time_slices: list[str],
        base_url: str | None = None
) -> list[str]:
    """Build and return a list of valid GCS public URLs."""
    if base_url is None:
        base_url = os.environ.get("NWM_BASE_URL", GOOGLE_CLOUD_BUCKET_URL)
    urls = []
    rd = reference_date.strftime("nwm.%Y%m%d/")
    for pf in prefixes:
        for ts in time_slices:
            urls.append(
                base_url +
                rd +
                configuration +
                pf +
                file_type +
                ts +
                suffix
                )
    return urls

def analysis_assim_extend_alaska_no_da(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for analysis_assim_extend_alaska_no_da.
    """
    configuration = "analysis_assim_extend_alaska_no_da/"
    prefixes = ["nwm.t20z."]
    file_type = "analysis_assim_extend_no_da.channel_rt."
    suffix = "alaska.nc"
    time_slices = ["tm" + str(t).zfill(2) + "." for t in range(8, 32)]
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def analysis_assim_extend_no_da(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for analysis_assim_extend_no_da.
    """
    configuration = "analysis_assim_extend_no_da/"
    prefixes = ["nwm.t16z."]
    file_type = "analysis_assim_extend_no_da.channel_rt."
    suffix = "conus.nc"
    time_slices = ["tm" + str(t).zfill(2) + "." for t in range(4, 28)]
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def analysis_assim_hawaii_no_da(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for analysis_assim_hawaii_no_da.
    """
    configuration = "analysis_assim_hawaii_no_da/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24)]
    file_type = "analysis_assim_no_da.channel_rt."
    suffix = "hawaii.nc"
    time_slices = ["tm" + str(t).zfill(4) + "." for t in range(200, 260, 15)]
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def analysis_assim_puerto_rico_no_da(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for analysis_assim_puertorico_no_da.
    """
    configuration = "analysis_assim_puertorico_no_da/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24)]
    file_type = "analysis_assim_no_da.channel_rt."
    suffix = "puertorico.nc"
    time_slices = ["tm02."]
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def medium_range_mem_1(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for medium_range_mem1.
    """
    configuration = "medium_range_mem1/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    file_type = "medium_range.channel_rt_1."
    suffix = "conus.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 241)]
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def medium_range_blend(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for medium_range_blend.
    """
    configuration = "medium_range_blend/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    file_type = "medium_range_blend.channel_rt."
    suffix = "conus.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 241)]
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def medium_range_no_da(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for medium_range_no_da.
    """
    configuration = "medium_range_no_da/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    file_type = "medium_range_no_da.channel_rt."
    suffix = "conus.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(3, 241, 3)]
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def medium_range_alaska_mem_1(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for medium_range_alaska_mem1.
    """
    configuration = "medium_range_alaska_mem1/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    file_type = "medium_range.channel_rt_1."
    suffix = "alaska.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 241)]
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def medium_range_blend_alaska(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for medium_range_blend_alaska.
    """
    configuration = "medium_range_blend_alaska/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    file_type = "medium_range_blend.channel_rt."
    suffix = "alaska.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 241)]
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def medium_range_alaska_no_da(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for medium_range_alaska_no_da.
    """
    configuration = "medium_range_alaska_no_da/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    file_type = "medium_range_no_da.channel_rt."
    suffix = "alaska.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(3, 241, 3)]
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def short_range(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for short_range.
    """
    configuration = "short_range/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24)]
    file_type = "short_range.channel_rt."
    suffix = "conus.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 19)]
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def short_range_alaska(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for short_range_alaska.
    """
    configuration = "short_range_alaska/"
    prefixes_15 = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    prefixes_45 = ["nwm.t" + str(p).zfill(2) + "z." for p in range(3, 27, 6)]
    file_type = "short_range.channel_rt."
    suffix = "alaska.nc"
    time_slices_15 = ["f" + str(p).zfill(3) + "." for p in range(1, 16)]
    time_slices_45 = ["f" + str(p).zfill(3) + "." for p in range(1, 46)]
    urls_15 = build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes_15,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices_15
    )
    urls_45 = build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes_45,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices_45
    )
    return urls_15 + urls_45

def short_range_hawaii(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for short_range_hawaii.
    """
    configuration = "short_range_hawaii/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 12)]
    file_type = "short_range.channel_rt."
    suffix = "hawaii.nc"
    time_slices = []
    for h in range(0, 4900, 100):
        for m in range(0, 60, 15):
            time_slices.append("f" + str(h+m).zfill(5) + ".")
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices[1:-3]
    )

def short_range_hawaii_no_da(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for short_range_hawaii_no_da.
    """
    configuration = "short_range_hawaii_no_da/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 12)]
    file_type = "short_range_no_da.channel_rt."
    suffix = "hawaii.nc"
    time_slices = []
    for h in range(0, 4900, 100):
        for m in range(0, 60, 15):
            time_slices.append("f" + str(h+m).zfill(5) + ".")
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices[1:-3]
    )

def short_range_puerto_rico(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for short_range_puertorico.
    """
    configuration = "short_range_puertorico/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(6, 30, 12)]
    file_type = "short_range.channel_rt."
    suffix = "puertorico.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 49)]
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def short_range_puerto_rico_no_da(
        reference_date: pd.Timestamp
) -> list[str]:
    """
    Generate public urls for short_range_puertorico_no_da.
    """
    configuration = "short_range_puertorico_no_da/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(6, 30, 12)]
    file_type = "short_range_no_da.channel_rt."
    suffix = "puertorico.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 49)]
    return build_gcs_public_urls(
        reference_date=reference_date,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

NWM_URL_BUILDERS: dict[tuple[ModelDomain, ModelConfiguration], URLBuilder] = {
    (ModelDomain.ALASKA,
     ModelConfiguration.ANALYSIS_ASSIM_EXTEND_ALASKA_NO_DA): analysis_assim_extend_alaska_no_da,
    (ModelDomain.CONUS,
     ModelConfiguration.ANALYSIS_ASSIM_EXTEND_NO_DA): analysis_assim_extend_no_da,
    (ModelDomain.HAWAII,
     ModelConfiguration.ANALYSIS_ASSIM_HAWAII_NO_DA): analysis_assim_hawaii_no_da,
    (ModelDomain.PUERTO_RICO,
     ModelConfiguration.ANALYSIS_ASSIM_PUERTO_RICO_NO_DA): analysis_assim_puerto_rico_no_da,
    (ModelDomain.CONUS, ModelConfiguration.MEDIUM_RANGE_MEM_1): medium_range_mem_1,
    (ModelDomain.CONUS, ModelConfiguration.MEDIUM_RANGE_BLEND): medium_range_blend,
    (ModelDomain.CONUS, ModelConfiguration.MEDIUM_RANGE_NO_DA): medium_range_no_da,
    (ModelDomain.ALASKA, ModelConfiguration.MEDIUM_RANGE_ALASKA_MEM_1): medium_range_alaska_mem_1,
    (ModelDomain.ALASKA, ModelConfiguration.MEDIUM_RANGE_BLEND_ALASKA): medium_range_blend_alaska,
    (ModelDomain.ALASKA, ModelConfiguration.MEDIUM_RANGE_ALASKA_NO_DA): medium_range_alaska_no_da,
    (ModelDomain.CONUS, ModelConfiguration.SHORT_RANGE): short_range,
    (ModelDomain.ALASKA, ModelConfiguration.SHORT_RANGE_ALASKA): short_range_alaska,
    (ModelDomain.HAWAII, ModelConfiguration.SHORT_RANGE_HAWAII): short_range_hawaii,
    (ModelDomain.HAWAII, ModelConfiguration.SHORT_RANGE_HAWAII_NO_DA): short_range_hawaii_no_da,
    (ModelDomain.PUERTO_RICO,
     ModelConfiguration.SHORT_RANGE_PUERTO_RICO): short_range_puerto_rico,
    (ModelDomain.PUERTO_RICO,
     ModelConfiguration.SHORT_RANGE_PUERTO_RICO_NO_DA): short_range_puerto_rico_no_da
}
"""Mapping from (ModelDomain, ModelConfiguration) to url builder function."""

def build_nwm_filepath(
    root: Path,
    configuration: ModelConfiguration,
    reference_date: pd.Timestamp
    ) -> Path:
    """Build and return NWM hive-partitioned parquet file path."""
    config = f"configuration={configuration}"
    year = f"year={reference_date.year}"
    month = f"month={reference_date.month}"
    day = f"D{reference_date.day}.parquet"
    subdirectory = SUBDIRECTORIES["nwm"]
    return root / subdirectory / config / year / month / day

def download_nwm(
        start: pd.Timestamp,
        end: pd.Timestamp,
        root: Path,
        jobs: int = 1,
        retries: int = 3
) -> None:
    """
    Download and process NWM output.

    Parameters
    ----------
    start: pandas.Timestamp
        First reference date to retrieve and process.
    end: pandas.Timestamp
        Last refrence date to retrieve and process.
    root: pathlib.Path
        Root data directory.
    jobs: int, optional, default 1
        Number of parallel process used to process NWM output.
    retries: int, optional, default 3
        Number of times to retry NetCDF file downloads.
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # Load routelink
    routelink = download_routelink(root).select(
        ["nwm_feature_id", "domain"]
    ).collect()

    # Generate reference dates
    logger.info("Generating reference dates")
    reference_dates = generate_reference_dates(start, end)

    # Temporary directory
    odir = root / "temp"
    odir.mkdir(exist_ok=True)
    logger.info("Downloading to %s", odir)

    # Download and process data
    features = {}
    for (domain, configuration), builder in NWM_URL_BUILDERS.items():
        # Extract features
        if domain not in features:
            features[domain] = routelink.filter(
                pl.col("domain") == domain)["nwm_feature_id"].to_numpy()

        # Process each reference day
        for rd in reference_dates:
            # Output file
            ofile = build_nwm_filepath(root, configuration, rd)

            # Skip existing files
            if ofile.exists():
                logger.info("Skipping existing file %s", ofile)
                continue

            logger.info("Building %s", ofile)
            ofile.parent.mkdir(exist_ok=True, parents=True)

            logger.info("Generating Google Cloud URLs")
            urls = builder(rd)

            # Temporary download directory
            with TemporaryDirectory(
                prefix=f"{configuration}_{rd.strftime("%Y%m%d")}_",
                dir=odir
                ) as td:
                file_paths = [Path(td) / f"part_{i}.nc" for i in range(len(urls))]

                logger.info("Downloading NWM data")
                download_files(
                    *list(zip(urls, file_paths)),
                    timeout=3600,
                    file_validator=netcdf_validator,
                    retries=retries,
                    limit=20
                )

                logger.info("Verifying files")
                verified = [f for f in file_paths if f.exists()]
                if not verified:
                    warnings.warn("No files verified", RuntimeWarning)
                    continue

                logger.info("Processing NWM data")
                data = process_netcdf_parallel(
                    verified,
                    ["streamflow"],
                    features[domain],
                    jobs
                )
                logger.info("Saving %s", ofile)
                pl.DataFrame(data).with_columns(
                    pl.col("value_time").dt.cast_time_unit("ms"),
                    pl.col("reference_time").dt.cast_time_unit("ms")
                ).write_parquet(ofile)

def scan_nwm_no_cache(root: Path) -> pl.LazyFrame:
    """
    Return polars.LazyFrame of NWM output.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # Generate reference dates
    logger.info("Scanning predictions")
    subdirectory = SUBDIRECTORIES["nwm"]
    return pl.scan_parquet(
        root / f"{subdirectory}/",
        hive_schema={
            "configuration": pl.Enum(ModelConfiguration),
            "year": pl.Int32,
            "month": pl.Int32
        }
    )

@functools.lru_cache(LRU_CACHE_SIZES["nwm"])
def scan_nwm_cache(root: Path) -> pl.LazyFrame:
    """
    Return polars.LazyFrame of NWM output. Cache LazyFrame.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    """
    return scan_nwm_no_cache(root)

def scan_nwm(root: Path, cache: bool = False) -> pl.LazyFrame:
    """
    Return polars.LazyFrame of NWM output.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    cache: bool, optional, default False
        If True, cache underlying LazyFrame.
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    logger.info("Scanning predictions")
    if cache:
        return scan_nwm_cache(root)
    return scan_nwm_no_cache(root)

def load_nwm_no_cache(
    root: Path,
    configuration: ModelConfiguration,
    year: int,
    month: int,
    nwm_feature_id: int | None = None
    ) -> pl.DataFrame:
    """
    Return polars.DataFrame of USGS observations.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    configuration: ModelConfiguration
        Model configuration (e.g. ModelConfiguration.MEDIUM_RANGE_MEM_1)
    year: int
        Integer year.
    month: int
        Integer month.
    nwm_feature_id: int, optional
        NWM channel feature ID.
    
    Returns
    -------
    polars.DataFrame
    """
    # Initial scan
    df = scan_nwm(root).filter(
        pl.col("configuration") == configuration,
        pl.col("year") == year,
        pl.col("month") == month
    ).select(
        ["nwm_feature_id", "reference_time", "value_time", "predicted_cfs"]
    )

    # Feature by filter
    if nwm_feature_id is not None:
        df = df.filter(pl.col("nwm_feature_id") == nwm_feature_id)
    return df.collect().sort(
        ["nwm_feature_id", "reference_time", "value_time"]
    )

@functools.lru_cache(LRU_CACHE_SIZES["nwm"])
def load_nwm_cache(
    root: Path,
    configuration: ModelConfiguration,
    year: int,
    month: int,
    nwm_feature_id: int | None = None
    ) -> pl.DataFrame:
    """
    Return polars.DataFrame of USGS observations. Cache result.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    configuration: ModelConfiguration
        Model configuration (e.g. ModelConfiguration.MEDIUM_RANGE_MEM_1)
    year: int
        Integer year.
    month: int
        Integer month.
    nwm_feature_id: int, optional
        NWM channel feature ID.
    
    Returns
    -------
    polars.DataFrame
    """
    # Initial scan with caching
    df = scan_nwm(root, cache=True).filter(
        pl.col("configuration") == configuration,
        pl.col("year") == year,
        pl.col("month") == month
    ).select(
        ["nwm_feature_id", "reference_time", "value_time", "predicted_cfs"]
    )

    # Feature by filter
    if nwm_feature_id is not None:
        df = df.filter(pl.col("nwm_feature_id") == nwm_feature_id)
    return df.collect().sort(
        ["nwm_feature_id", "reference_time", "value_time"]
    )

def load_nwm(
    root: Path,
    configuration: ModelConfiguration,
    year: int,
    month: int,
    cache: bool = False
    ) -> pl.DataFrame:
    """
    Return polars.DataFrame of NWM output.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    configuration: ModelConfiguration
        Model configuration (e.g. ModelConfiguration.MEDIUM_RANGE_MEM_1)
    year: int
        Integer year.
    month: int
        Integer month.
    cache: bool, optional, default False
        If True, cache underlying dataframe.
    
    Returns
    -------
    polars.DataFrame
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # Generate reference dates
    logger.info("Retrieving predictions")
    if cache:
        return load_nwm_cache(root, configuration, year, month)
    return load_nwm_no_cache(root, configuration, year, month)

def nwm_site_generator(
    root: Path,
    configuration: ModelConfiguration,
    nwm_feature_id: int,
    start_time: pd.Timestamp,
    end_time: pd.Timestamp,
    cache: bool = False
    ) -> Generator[pl.DataFrame]:
    """
    Iteratively return polars.DataFrame of NWM output.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    configuration: ModelConfiguration
        Model configuration (e.g. ModelConfiguration.MEDIUM_RANGE_MEM_1)
    nwm_feature_id: str
        NWM channel feature ID.
    start_time: pandas.Timestamp
        Earliest reference time to retrieve.
    end_time: pandas.Timestamp
        Latest reference time to retrieve.
    cache: bool, optional, default False
        Whether to cache the underlying DataFrames for subsequent calls.
    
    Returns
    -------
    polars.DataFrame
    """
    # Pad retrieval
    padded_start_time = start_time - pd.Timedelta("31D")
    padded_end_time = end_time + pd.Timedelta("31D")

    # Load data
    for m in pd.date_range(padded_start_time, padded_end_time, freq="1ME"):
        # Get month of data
        if cache:
            df = load_nwm_cache(root, configuration, m.year, m.month, nwm_feature_id)
        else:
            df = load_nwm_no_cache(root, configuration, m.year, m.month, nwm_feature_id)

        yield df.filter(
            pl.col("reference_time") >= start_time,
            pl.col("reference_time") <= end_time
        ).unique(["reference_time", "value_time"]).sort(["reference_time", "value_time"])

def load_nwm_site(
    root: Path,
    configuration: ModelConfiguration,
    nwm_feature_id: int,
    start_time: pd.Timestamp,
    end_time: pd.Timestamp,
    cache: bool = False
    ) -> pl.DataFrame:
    """
    Return polars.DataFrame of NWM output.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    configuration: ModelConfiguration
        Model configuration (e.g. ModelConfiguration.MEDIUM_RANGE_MEM_1)
    nwm_feature_id: str
        NWM channel feature ID.
    start_time: pandas.Timestamp
        Earliest reference time to retrieve.
    end_time: pandas.Timestamp
        Latest reference time to retrieve.
    cache: bool, optional, default False
        Whether to cache the underlying DataFrames for subsequent calls.
    
    Returns
    -------
    polars.DataFrame
    """
    # Collect dataframes
    dataframes = []
    for df in nwm_site_generator(
        root=root,
        configuration=configuration,
        nwm_feature_id=nwm_feature_id,
        start_time=start_time,
        end_time=end_time,
        cache=cache
    ):
        dataframes.append(df)
    return pl.concat(dataframes)



"""Download and process RouteLink information."""
from pathlib import Path
from tempfile import TemporaryDirectory
import inspect
import tarfile

import pandas as pd
import polars as pl
from yarl import URL

from nwm_explorer.logger import get_logger
from nwm_explorer.downloads import download_files
from nwm_explorer.constants import (ModelDomain, ROUTELINK_URL, ROUTELINK_PARQUET,
    ROUTELINK_FILENAMES)

def download_routelink(
        root: Path,
        url: str | URL = ROUTELINK_URL
) -> pl.LazyFrame:
    """
    Download RouteLink file.

    Parameters
    ----------
    root: pathlib.Path
        Root data directory.
    url: str | URL
        Source URL.
    
    Returns
    -------
    polars.LazyFrame
    """
    # Get logger
    name = __loader__.name + "." + inspect.currentframe().f_code.co_name
    logger = get_logger(name)

    # Check for file
    file_path = root / ROUTELINK_PARQUET
    if file_path.exists():
        logger.info("Scanning %s", file_path)
        return pl.scan_parquet(file_path)
    logger.info("Downloading %s", file_path)

    # Download RouteLink
    with TemporaryDirectory() as td:
        # Temporary download path
        ofile = Path(td) / "routelink.tar.gz"

        # Download
        download_files(
            (url, ofile),
        )

        logger.info("Extracting routelink files")
        odir = Path(td) / "routelinks"
        odir.mkdir()
        with tarfile.open(ofile, "r:gz") as tf:
            tf.extractall(odir)

        logger.info("Processing routelink files")
        dfs = []
        for d, fn in ROUTELINK_FILENAMES.items():
            ifile = odir / f"csv/{fn}"
            df = pd.read_csv(
                ifile,
                comment="#",
                dtype=str
            )
            df["domain"] = d
            dfs.append(df)

        # Clean-up
        data = pd.concat(dfs, ignore_index=True)
        data = data[data["usgs_site_code"].str.isdigit()]
        short = data["usgs_site_code"].str.len() <= 7
        data.loc[short, "usgs_site_code"] = "0" + data.loc[short, "usgs_site_code"]

        # Save
        pl_data = pl.DataFrame(
            data,
            schema_overrides={
                "usgs_site_code": pl.String,
                "domain": ModelDomain,
                "nwm_feature_id": pl.Int64,
                "latitude": pl.Float64,
                "longitude": pl.Float64
                },
            strict=False
        ).drop_nulls("usgs_site_code")
        pl_data.write_parquet(file_path)

    # Scan
    logger.info("Scanning %s", file_path)
    return pl.scan_parquet(file_path)



"""Methods to generate views of data."""
from typing import Callable, Any, Generator
from itertools import cycle, count
from dataclasses import dataclass

import polars as pl
import panel as pn
from panel.viewable import Viewer
import plotly.graph_objects as go
from plotly.colors import hex_to_rgb
import colorcet as cc
import pandas as pd
import numpy.typing as npt
import numpy as np
import geopandas as gpd

from nwm_explorer.nwm import ModelConfiguration
from nwm_explorer.evaluate import Metric
from nwm_explorer.constants import (ModelDomain, GROUP_SPECIFICATIONS, DOMAIN_LOOKUP,
    CONFIGURATION_LOOKUP, NWMGroupSpecification, METRIC_LOOKUP, RANK_LOOKUP,
    PlotlyFigure, DEFAULT_CENTER, DEFAULT_ZOOM, AxisType)
from nwm_explorer.configuration import MapLayer

pn.extension('tabulator')

def build_lead_time_lookup(
        group_specificiations: dict[ModelConfiguration, NWMGroupSpecification] | None = None
) -> dict[ModelConfiguration, list[int]]:
    """build."""
    # Check specifications
    if group_specificiations is None:
        group_specificiations = GROUP_SPECIFICATIONS

    # Build lead time lists
    lead_time_lookup = {}
    for c, s in group_specificiations.items():
        lead_time_lookup[c] = list(
            range(0, s.lead_time_hours_max+s.window_interval, s.window_interval)
        )
    return lead_time_lookup

def build_lead_time_strings(
        lead_time_lookup: dict[ModelConfiguration, list[int]] | None = None
) -> dict[ModelConfiguration, list[str]]:
    """build."""
    # Check specifications
    if lead_time_lookup is None:
        lead_time_lookup = build_lead_time_lookup()

    # Build lead time lists
    lead_time_strings = {}
    for c, lead_times in lead_time_lookup.items():
        if len(lead_times) == 1:
            lead_time_strings[c] = [str(l) for l in lead_times]
            continue

        lead_time_strings[c] = []
        for idx in range(0, len(lead_times)-1):
            lead_time_strings[c].append(f"{lead_times[idx]} to <{lead_times[idx+1]}")
        lead_time_strings[c].append(f"{lead_times[-1]}")

    return lead_time_strings

class FilterWidgets(Viewer):
    """Holds various data filtering widgets and values."""
    def __init__(self, evaluation_options: list[str], **params):
        super().__init__(**params)

        # Merge domain and configuration look-ups
        self._configuration_lookup = {
            DOMAIN_LOOKUP[k]+v: (DOMAIN_LOOKUP[k], k) for k, v in CONFIGURATION_LOOKUP.items()
            }

        # Setup widgets
        self._widgets: dict[str, pn.widgets.Widget] = {
            "label": pn.widgets.Select(
                name="Evaluation",
                options=evaluation_options
            ),
            "configuration": pn.widgets.Select(
                name="Model Configuration",
                options=list(self._configuration_lookup.keys())
            ),
            "metric": pn.widgets.Select(
                name="Metric",
                options=list(METRIC_LOOKUP.keys())
            ),
            "rank": pn.widgets.Select(
                name="Streamflow aggregation method",
                options=list(RANK_LOOKUP.keys())
            ),
            "significant": pn.widgets.Select(
                name="Show",
                options=["All sites", "Statistically significant"]
            ),
            "lead_time": pn.pane.Placeholder()
        }

        # Generate lead times
        self._lead_time_lookup = build_lead_time_lookup()
        self._lead_time_strings = build_lead_time_strings()

        # Add lead time widget
        self._widgets["lead_time"].object = pn.widgets.DiscretePlayer(
            name="Lead time (hours)",
            options=self._lead_time_strings[self.configuration],
            show_loop_controls=False,
            visible_buttons=["previous", "next"],
            width=300,
            value=self._lead_time_strings[self.configuration][0]
            )

        # Keep track of callbacks
        self._callbacks: list[Callable] = []

        # Update lead time
        def update_lead_times(event, callback_type: str) -> None:
            # Ignore non-events
            if event is None or callback_type is None:
                return

            # Maintain value
            value = self._widgets["lead_time"].object.value
            if value not in self._lead_time_strings[self.configuration]:
                value = self._lead_time_strings[self.configuration][0]

            # Create new widget to force refresh
            self._widgets["lead_time"].object = pn.widgets.DiscretePlayer(
                name="Lead time (hours)",
                options=self._lead_time_strings[self.configuration],
                show_loop_controls=False,
                visible_buttons=["previous", "next"],
                width=300,
                value=value
                )

            # Bind callbacks
            for c in self._callbacks:
                pn.bind(c, self._widgets["lead_time"].object.param.value, watch=True,
                    callback_type="lead_time")
        pn.bind(update_lead_times, self._widgets["configuration"].param.value, watch=True,
            callback_type="configuration")

        # Create layout
        self._layout = pn.Card(
            pn.Column(*list(self._widgets.values())),
            title="Filters",
            collapsible=False,
            width=350
            )

    def __panel__(self):
        return self._layout

    def bind(self, function) -> None:
        """Bind a function to all filtering widgets."""
        self._callbacks.append(function)
        for k, v in self._widgets.items():
            if isinstance(v, pn.pane.Placeholder):
                pn.bind(function, v.object.param.value, watch=True, callback_type=k)
            else:
                pn.bind(function, v.param.value, watch=True, callback_type=k)

    @property
    def label(self) -> str:
        """Currently selected evaluation label."""
        return self._widgets["label"].value

    @property
    def configuration(self) -> ModelConfiguration:
        """Currently selected model configuration."""
        return self._configuration_lookup[self._widgets["configuration"].value][1]

    @property
    def metric(self) -> Metric:
        """Currently selected evaluation metric."""
        return METRIC_LOOKUP[self._widgets["metric"].value]

    @property
    def metric_label(self) -> str:
        """Currently selected evaluation metric label."""
        return self._widgets["metric"].value

    @property
    def rank(self) -> str:
        """Currently selected streamflow aggregation method."""
        return RANK_LOOKUP[self._widgets["rank"].value]

    @property
    def significant(self) -> bool:
        """Only show 'statistically significant sites."""
        return self._widgets["significant"].value == "Statistically significant"

    @property
    def point_column(self) -> str:
        """Construct a metrics dataframe column label from current selection."""
        return self.metric + "_" + self.rank + "_point"

    @property
    def upper_column(self) -> str:
        """Construct a metrics dataframe column label from current selection."""
        return self.metric + "_" + self.rank + "_upper"

    @property
    def lower_column(self) -> str:
        """Construct a metrics dataframe column label from current selection."""
        return self.metric + "_" + self.rank + "_lower"

    @property
    def domain(self) -> ModelDomain:
        """Currently selected domain."""
        return self._configuration_lookup[self._widgets["configuration"].value][0]

    @property
    def lead_time(self) -> int:
        """Currently selected minimum lead time in hours."""
        lead_time_string = self._widgets["lead_time"].object.value
        idx = self._lead_time_strings[self.configuration].index(lead_time_string)
        return self._lead_time_lookup[self.configuration][idx]

class TableView(Viewer):
    """Handles tabular view of data."""
    def __init__(self, **params):
        super().__init__(**params)

        # Setup placeholder
        self._table = pn.pane.Placeholder()

        # Create layout
        self._layout = self._table

    def __panel__(self):
        return pn.Card(
            self._layout,
            collapsible=False,
            hide_header=True
            )

    def update(self, dataframe: pl.DataFrame) -> None:
        """Update data underlying tabular representation."""
        self._table.object = pn.widgets.Tabulator(
            dataframe.to_pandas(),
            height=500,
            show_index=False
        )

class MapView(Viewer):
    """Display data on a map."""
    def __init__(self, map_layers: list[MapLayer] | None = None, **params):
        super().__init__(**params)

        # Setup extra layers
        index = count(2)
        self.layers = {l.name: (next(index), l) for l in map_layers}
        self.layer_selector = pn.widgets.CheckBoxGroup(
            name="Map Layers",
            options=[l.name for l in map_layers]
        )
        self.map_layer_selector = pn.Column(
            "# Map Layers",
            self.layer_selector
        )

        # Stub layers
        layer_color = cycle(cc.glasbey_light)
        layer_data = [
            go.Scattermap(
                    showlegend=False,
                    name="",
                    mode="markers",
                    marker=dict(size=10, color=next(layer_color))
                ) for _ in map_layers
        ]

        # Setup main map
        self._domain = None
        self._figure = PlotlyFigure(
            data=[
                go.Scattermap(
                    showlegend=False,
                    name="",
                    mode="markers",
                    marker=dict(size=20, color="#00FFF2")
                ),
                go.Scattermap(
                    marker=dict(
                        colorbar=dict(
                            title=dict(
                                side="right"
                                )
                            ),
                        size=15,
                        colorscale=cc.CET_L17
                    ),
                    showlegend=False,
                    name="",
                    mode="markers"
                    )
                ] + layer_data,
            layout=go.Layout(
                showlegend=False,
                height=540,
                width=850,
                margin=dict(l=0, r=0, t=0, b=0),
                map=dict(
                    style="satellite-streets"
                ),
                clickmode="event",
                modebar=dict(
                    remove=["lasso", "select", "resetview"],
                    orientation="v"
                ),
                dragmode="zoom"
            )
        )

        # Create layout
        self._pane = pn.pane.Plotly(self._figure, config={'displaylogo': False})

        # Track selection state
        self.click_data: dict[str, Any] = {}
        def catch_click(event, callback_type: str) -> None:
            if event is None or callback_type is None:
                return

            # Handle selection/deselection
            lat, lon = event["points"][0].get("lat"), event["points"][0].get("lon")
            if (self.click_data.get("lat") == lat) & (self.click_data.get("lon") == lon):
                self.click_data = {}
            else:
                self.click_data.update(event["points"][0])

            # Update marker
            self.update_marker()
        pn.bind(catch_click, self._pane.param.click_data, watch=True, callback_type="click")

        # Viewport state
        self.viewport: dict[str, float] = {}
        def catch_relayout(event: dict[str, Any], callback_type: str) -> None:
            if event is None or callback_type is None:
                return

            # Check for viewport change
            bbox = event.get("map._derived")
            if bbox is None:
                return

            # Extract points
            self.viewport = {
                "lat_max": bbox["coordinates"][0][1],
                "lat_min": bbox["coordinates"][2][1],
                "lon_max": bbox["coordinates"][1][0],
                "lon_min": bbox["coordinates"][0][0]
            }
        pn.bind(catch_relayout, self._pane.param.relayout_data, watch=True,
            callback_type="relayout")

        # Layer selection
        def handle_layer_selection(event: list[str], callback_type: str) -> None:
            if event is None or callback_type is None:
                return

            # Select/deselect each layer
            for name, (idx, layer) in self.layers.items():
                if name in event and self._figure["data"][idx]["lat"] is None:
                    # Load data
                    if layer.columns is None:
                        cols = ["geometry"]
                    elif "geometry" not in layer.columns:
                        cols = layer.columns+["geometry"]
                    else:
                        cols = layer.columns
                    gdf = gpd.read_parquet(layer.path, columns=cols)

                    # Custom data
                    if len(cols) != 1:
                        custom_data = gdf[layer.columns]
                    else:
                        custom_data = None

                    # Hover template
                    hover_template = ""
                    for cindex, cname in enumerate(layer.columns):
                        hover_template += f"{cname}: " + "%{customdata[" + str(cindex) + "]}<br>"
                    hover_template += "Longitude: %{lon}<br>Latitude: %{lat}"

                    # Update trace
                    self._figure["data"][idx].update(
                        lat=gdf["geometry"].y,
                        lon=gdf["geometry"].x,
                        customdata=custom_data,
                        hovertemplate=hover_template
                    )
                else:
                    self._figure["data"][idx].update(
                        lat=None,
                        lon=None,
                        customdata=None,
                        hovertemplate=None
                    )

            # Update
            self._pane.object = self._figure
        pn.bind(handle_layer_selection, self.layer_selector.param.value, watch=True,
            callback_type="layer")

    def __panel__(self):
        return pn.Card(
            self._pane,
            collapsible=False,
            hide_header=True
            )

    def update_marker(self) -> None:
        """Move or remove selection marker."""
        self._figure["data"][0].update(
            lat=[self.click_data.get("lat")],
            lon=[self.click_data.get("lon")]
        )
        self._pane.object = self._figure

    def bind_click(self, function) -> None:
        """Bind a function to click event."""
        pn.bind(function, self._pane.param.click_data, watch=True, callback_type="click")

    def bind_relayout(self, function) -> None:
        """Bind a function to relayout event."""
        pn.bind(function, self._pane.param.relayout_data, watch=True, callback_type="relayout")

    def update(
            self,
            dataframe: pl.DataFrame,
            column: str,
            domain: ModelDomain,
            cmin: float,
            cmax: float,
            metric_label: str,
            custom_data: pd.DataFrame | None = None,
            hover_template: str | None = None
            ) -> None:
        """Update map."""
        # Update domain
        if self._domain != domain:
            self._domain = domain
            self.viewport = {}

        # Handle selection
        show_marker = self.click_data.get("lat") in dataframe["latitude"]
        show_marker &= self.click_data.get("lon") in dataframe["longitude"]
        if not show_marker:
            self.click_data = {}
            self.update_marker()

        # Update markers
        self._figure["data"][1]["marker"].update(
            color=dataframe[column],
            cmin=cmin,
            cmax=cmax
        )
        self._figure["data"][1].update(
            lat=dataframe["latitude"],
            lon=dataframe["longitude"],
            customdata=custom_data,
            hovertemplate=hover_template
        )
        self._figure["data"][1]["marker"]["colorbar"]["title"].update(text=metric_label)

        # Update focus
        self._figure["layout"].update(
            uirevision=domain,
            selectionrevision=(self.click_data.get("lat"), self.click_data.get("lon"))
        )
        self._figure["layout"]["map"].update(
            center=DEFAULT_CENTER[domain],
            zoom=DEFAULT_ZOOM[domain]
        )

        # Send updates to frontend
        self._pane.object = self._figure

def rgb_to_hex(rgb_tuple):
    """Converts an RGB tuple to a hexadecimal color string."""
    r, g, b = rgb_tuple
    return f"#{r:02x}{g:02x}{b:02x}"

def invert_color(value: str) -> str:
    """Convert a hex color to an inverted hex color.
    
    Parameters
    ----------
    value: str, required,
        Hex color string.
    
    Returns
    -------
    str:
        Inverted hex color.
    """
    r, g, b = hex_to_rgb(value)
    return rgb_to_hex((255-r, 255-g, 255-b))

class TimeSeriesView(Viewer):
    """Display time series data."""
    def __init__(self, **params):
        super().__init__(**params)

        # Setup
        self._figure = PlotlyFigure(
            data=[go.Scatter()],
            layout=go.Layout(
                height=250,
                width=1045,
                margin=dict(l=0, r=25, t=10, b=0),
                yaxis=dict(title=dict(text="Streamflow (CFS)")),
                clickmode="event",
                modebar=dict(
                    remove=["resetview"],
                    orientation="v"
                ),
                showlegend=False
            )
        )

        # Create layout
        self._pane = pn.pane.Plotly(self._figure, config={'displaylogo': False})

        # Color ramp
        self._color_ramp = cycle(cc.CET_C7)

        # Track selection state
        self.click_data: dict[str, Any] = {}
        self.selected_traces: dict[int, bool] = {}
        def catch_click(event, callback_type: str) -> None:
            if event is None or callback_type is None:
                return

            # Get curve number
            curve_number = event["points"][0].get("curveNumber")

            # Highlight/unhighlight trace
            self.toggle_highlight(curve_number)

            # Update click data
            if self.click_data.get("curveNumber") == curve_number:
                self.click_data = {}
            else:
                self.click_data.update(event["points"][0])
        pn.bind(catch_click, self._pane.param.click_data, watch=True, callback_type="click")

    def __panel__(self):
        return pn.Card(
            self._pane,
            collapsible=False,
            hide_header=True
            )

    def set_yaxis_title(self, title: str) -> None:
        """Change y-axis title."""
        # Update
        self._figure["layout"]["yaxis"]["title"].update(text=title)

        # Refresh
        self._pane.object = self._figure

    def set_axis_type(self, axis_type: AxisType) -> None:
        """Change y-axis type."""
        # Update
        self._figure["layout"]["yaxis"].update(type=axis_type)

        # Refresh
        self._pane.object = self._figure

    def toggle_highlight(self, curve_number: int) -> None:
        """Highlight or unhighlight selected trace."""
        # Invert colors
        color = self._figure["data"][curve_number]["line"]["color"]
        self._figure["data"][curve_number]["line"].update(
            color=invert_color(color)
        )

        # Update width
        width = self._figure["data"][curve_number]["line"]["width"]
        if curve_number in self.selected_traces:
            # Remove trace
            self.selected_traces.pop(curve_number)

            # Reduce width
            self._figure["data"][curve_number]["line"].update(
                width=width-4
            )
        else:
            # Add trace
            self.selected_traces[curve_number] = True

            # Increase width
            self._figure["data"][curve_number]["line"].update(
                width=width+4
            )

        # Refresh
        self._pane.object = self._figure

    def erase_data(
            self,
            xrange: tuple[pd.Timestamp, pd.Timestamp] | None = None
        ) -> None:
        """Completely erase displayed time series."""
        # Overwrite old traces
        self._figure["data"] = [go.Scatter(
            mode="lines",
            line={"color": "#000000", "width": 4.0},
            name=""
        )]

        # Update x-range
        if xrange is not None:
            self._figure["layout"]["xaxis"].update(range=xrange)

        # Update focus
        self._figure["layout"].update(
            uirevision=pd.Timestamp.now().strftime("%Y%m%dT%H%M%S")
        )

        # Refresh
        self._pane.object = self._figure

        # Clear click data
        self.click_data = {}
        self.selected_traces = {}

    def update_trace(
            self,
            xdata: npt.ArrayLike,
            ydata: npt.ArrayLike,
            index: int = 0,
            name: str | None = None,
        ) -> None:
        """Update specific time series."""
        # Y-axis label
        ylabel = self._figure["layout"]["yaxis"]["title"]["text"]

        # Update x-y data
        self._figure["data"][index].update(
            x=xdata,
            y=ydata,
            hovertemplate=(
                f"<b>{name}</b><br>"
                "Valid: %{x}<br>"
                f"{ylabel}: " + "%{y:.2f}"
                )
        )

        # Refresh
        self._pane.object = self._figure

    def append_traces(
            self,
            traces: list[tuple[npt.ArrayLike, npt.ArrayLike, str]],
            mode: str = "lines"
        ) -> None:
        """Add time series."""
        # Y-axis label
        ylabel = self._figure["layout"]["yaxis"]["title"]["text"]

        # Add new traces
        for trace in traces:
            self._figure["data"].append(go.Scatter(
                x=trace[0],
                y=trace[1],
                mode=mode,
                line={"color": next(self._color_ramp), "width": 2.0},
                name="",
                hovertemplate=(
                    f"<b>{trace[2]}</b><br>"
                    "Valid: %{x}<br>"
                    f"{ylabel}: " + "%{y:.2f}"
                    )
            ))

        # Refresh
        self._pane.object = self._figure

class BarPlot(Viewer):
    """Display barplots."""
    def __init__(self, **params) -> None:
        super().__init__(**params)

        # Setup
        self.data = [go.Bar(
            name=""
        )]
        self.layout = go.Layout(
            height=250,
            width=440,
            margin=dict(l=0, r=25, t=10, b=0),
            modebar=dict(
                remove=["lasso", "select", "pan", "autoscale", "zoomin", "zoomout"],
                orientation="v"
            )
        )
        self.figure = dict(data=self.data, layout=self.layout)
        self.pane = pn.pane.Plotly(self.figure, config={'displaylogo': False})

    def update(
            self,
            xdata: npt.ArrayLike,
            ydata: npt.ArrayLike,
            ydata_lower: npt.ArrayLike,
            ydata_upper: npt.ArrayLike,
            xlabel: str,
            ylabel: str
        ) -> None:
        """Update data."""
        # Construct custom data
        custom_data = np.hstack((ydata_lower[:, np.newaxis], ydata_upper[:, np.newaxis]))

        # Update trace
        self.data[0].update(dict(
            x=xdata,
            y=ydata,
            customdata=custom_data,
            hovertemplate=(
                f"{ylabel}: " + "%{y:.2f}<br>" +
                "95% CI: %{customdata[0]:.2f} -- %{customdata[1]:.2f}<br>" +
                f"{xlabel}: " + "%{x}"
            ),
            error_y=dict(
                type="data",
                array=ydata_upper - ydata,
                arrayminus=ydata - ydata_lower
            )
        ))

        # Update axes
        self.layout.update(dict(
            xaxis=dict(title=dict(text=xlabel)),
            yaxis=dict(title=dict(text=ylabel))
        ))

        # Update frontend
        self.figure.update(dict(data=self.data, layout=self.layout))
        self.pane.object = self.figure

    def erase(
            self
        ) -> None:
        """Erase data."""
        self.data = [go.Bar(
            name=""
        )]

        # Update frontend
        self.figure.update(dict(data=self.data, layout=self.layout))
        self.pane.object = self.figure

    def __panel__(self) -> pn.Card:
        return pn.Card(
            self.pane,
            collapsible=False,
            hide_header=True
        )

class ECDFPlot(Viewer):
    """Display empirical cumulative distribution curves."""
    def __init__(self, **params) -> None:
        super().__init__(**params)

        # Setup
        self.data = [go.Scatter(
            name="",
            mode="markers"
        )]
        self.layout = go.Layout(
            height=255,
            width=300,
            margin=dict(l=0, r=25, t=10, b=0),
            modebar=dict(
                remove=["lasso", "select"],
                orientation="v"
            )
        )
        self.figure = dict(data=self.data, layout=self.layout)
        self.pane = pn.pane.Plotly(self.figure, config={'displaylogo': False})

    def update(
            self,
            xdata: npt.ArrayLike,
            xdata_lower: npt.ArrayLike,
            xdata_upper: npt.ArrayLike,
            xlabel: str,
            ylabel: str = "Empirical CDF",
            xrange: tuple[float, float] | None = None
        ) -> None:
        """Update data. Assumes xdata are sorted."""
        # Set xrange
        if xrange is None:
            xrange = (-2, 2)

        # Construct custom data
        custom_data = np.hstack((xdata_lower[:, np.newaxis], xdata_upper[:, np.newaxis]))

        # Plotting position
        ydata = np.arange(len(xdata)) / (len(xdata) + 1)

        # Update trace
        self.data[0].update(dict(
            x=xdata,
            y=ydata,
            customdata=custom_data,
            hovertemplate=(
                f"{xlabel}: " + "%{x:.2f}<br>" +
                "95% CI: %{customdata[0]:.2f} to %{customdata[1]:.2f}<br>" +
                f"{ylabel}: " + "%{y:.2f}"
            )
        ))

        # Update axes
        self.layout.update(dict(
            xaxis=dict(
                title=dict(text=xlabel),
                range=(xrange[0], xrange[1]),
                tickmode="linear",
                tick0=xrange[0],
                dtick=(xrange[1]-xrange[0])/4,
                tickformat=".1f"
            ),
            yaxis=dict(
                title=dict(text=ylabel),
                range=(0, 1),
                tickmode="linear",
                tick0=0.0,
                dtick=0.2,
                tickformat=".1f"
            )
        ))

        # Update frontend
        self.figure.update(dict(data=self.data, layout=self.layout))
        self.pane.object = self.figure

    def erase(
            self
        ) -> None:
        """Erase data."""
        self.data = [go.Scatter(
            name="",
            mode="markers"
        )]

        # Update frontend
        self.figure.update(dict(data=self.data, layout=self.layout))
        self.pane.object = self.figure

    def __panel__(self) -> pn.Card:
        return pn.Card(
            self.pane,
            collapsible=False,
            hide_header=True
        )

class ECDFMatrix(Viewer):
    """Show a set of linked empirical CDF plots."""
    def __init__(self, nplots: int, ncols: int, **params):
        super().__init__(**params)

        # Setup
        self.ncols = ncols
        self.plots = [ECDFPlot() for _ in range(nplots)]

    def update(self, index: int, **kwargs) -> None:
        """Update a plot."""
        self.plots[index].update(**kwargs)

    def erase(self, index: int) -> None:
        """Erase a single plot."""
        self.plots[index].erase()

    def erase_all(self) -> None:
        """Erase all plots."""
        for p in self.plots:
            p.erase()

    def __panel__(self) -> pn.GridBox:
        return pn.GridBox(*self.plots, ncols=self.ncols)

@dataclass
class ECDFParameters:
    """Dataclass used to hold empirical CDF plotting parameters."""
    index: int
    metric: Metric
    metric_label: str
    point_column: str
    upper_column: str
    lower_column: str

class ECDFSelector(Viewer):
    """WidetBox with metric selectors corresponding to ECDF plots."""
    def __init__(self, nplots: int, filter_widgets: FilterWidgets, **params):
        super().__init__(**params)

        # Setup
        self._filter_widgets = filter_widgets
        metric_names = list(METRIC_LOOKUP.keys())
        self._widgets = [
            pn.widgets.Select(
                    name=f"Metric {idx+1}",
                    options=metric_names,
                    value=metric_names[idx]
                )
        for idx in range(nplots)]

    def __panel__(self):
        return pn.Column("# Empirical CDF", *self._widgets)

    def __iter__(self) -> Generator[ECDFParameters]:
        for idx, w in enumerate(self._widgets):
            metric = METRIC_LOOKUP[w.value]
            yield ECDFParameters(
                index=idx,
                metric=metric,
                metric_label=w.value,
                point_column=metric + "_" + self._filter_widgets.rank + "_point",
                upper_column=metric + "_" + self._filter_widgets.rank + "_upper",
                lower_column=metric + "_" + self._filter_widgets.rank + "_lower"
            )

    def bind(self, function: Callable) -> None:
        """Bind function to widgets."""
        for w in self._widgets:
            pn.bind(function, w.param.value, watch=True, callback_type="ecdf")

class MarkdownView(Viewer):
    """Display Markdown content."""
    def __init__(self, **params) -> None:
        super().__init__(**params)

        self._pane = pn.pane.Markdown(
            "| Site Information |  |  \n| :-- | :-- |  \n|  |  |"
        )

    def update(self, content: str) -> None:
        """Update Markdown content."""
        self._pane.object = content

    def erase(self) -> None:
        """Clear Markdown content."""
        self._pane.object = "| Site Information |  |  \n| :-- | :-- |  \n|  |  |"

    def __panel__(self) -> None:
        return pn.Card(
            self._pane,
            collapsible=False,
            hide_header=True,
            width=350
        )
